# -*- coding: utf8 -*-
from __future__ import division
from warnings import warn
from generic import generic
from FGAme.mathtools import Vec2, dot, ux2D
from FGAme.mathtools import shadow_x, shadow_y, area, center_of_mass, clip, pi
from FGAme.physics import Collision
from FGAme.physics import Circle, AABB, Poly, Rectangle

DEFAULT_DIRECTIONS = [ux2D.rotate(n * pi / 12) for n in
                      [0, 1, 2, 3, 4, 5, 7, 8, 9, 10, 11]]


class CollisionError(Exception):

    '''Declara que não existem colisões disponíveis para os dois tipos de
    objetos'''
    pass


@generic
def get_collision(A, B):
    '''Retorna um objeto de colisão caso ocorra uma colisão com o objeto
    other. Caso não haja colisão, retorna None.

    Esta função é implementada por multidispatch. As classes derivadas de
    PhysicsObject devem registrar explicitamente a colisão entre todos os pares
    suportados (ex.: Circle com Circle, Circle com AABB, etc). Caso não tenha
    nenhuma implementação registrada, então utiliza-se a lógica de AABB's.'''

    tA = type(A).__name__
    tB = type(B).__name__
    warn('no collision defined for: (%s, %s)' % (tA, tB))

    return collision_circle(A, B)

###############################################################################
#                      Colisões entre objetos do mesmo tipo
###############################################################################


@get_collision.overload([Circle, Circle])
def collision_circle(A, B):
    '''Testa a colisão pela distância dos centros'''

    rA = A.cbb_radius
    rB = B.cbb_radius
    normal = B.pos - A.pos
    distance = normal.norm()

    if distance < rA + rB:
        normal /= distance
        delta = rA + rB - distance
        pos = A.pos + (rA - delta / 2) * normal
        return Collision(A, B, pos=pos, normal=normal, delta=delta)
    else:
        return None


@get_collision.overload([AABB, AABB])
def collision_aabb(A, B):
    '''Retorna uma colisão com o objeto other considerando apenas a caixas
    de contorno alinhadas ao eixo.'''

    # Detecta colisão pelas sombras das caixas de contorno
    rx, ry = B._pos - A._pos
    shadowx = A._delta_x + B._delta_x - abs(rx)
    shadowy = A._delta_y + B._delta_y - abs(ry)
    if shadowx <= 0 or shadowy <= 0:
        return None

    # Calcula ponto de colisão
    x_col = max(A.xmin, B.xmin) + shadowx / 2.
    y_col = max(A.ymin, B.ymin) + shadowy / 2.
    pos_col = Vec2(x_col, y_col)

    # Define sinal dos vetores normais: colisões tipo PONG
    if shadowx > shadowy:
        n = Vec2(0, (1 if A.ymin < B.ymin else -1))
        delta = shadowy
    else:
        n = Vec2((1 if A.xmin < B.xmin else -1), 0)
        delta = shadowx

    return Collision(A, B, pos=pos_col, normal=n, delta=delta)


@get_collision.overload([Poly, Poly])
def collision_poly(A, B, directions=None):
    '''Implementa a colisão entre dois polígonos arbitrários'''

    # Cria a lista de direções a partir das normais do polígono
    if directions is None:
        if A.num_normals + B.num_normals < 9:
            directions = A.get_normals() + B.get_normals()
        else:
            directions = DEFAULT_DIRECTIONS

    # Testa se há superposição de sombras em todas as direções consideradas
    # e calcula o menor valor para sombra e a direção normal
    min_shadow = float('inf')
    norm = None
    for u in directions:
        A_coords = [round(dot(pt, u), 6) for pt in A.vertices]
        B_coords = [round(dot(pt, u), 6) for pt in B.vertices]
        Amax, Amin = max(A_coords), min(A_coords)
        Bmax, Bmin = max(B_coords), min(B_coords)
        minmax, maxmin = min(Amax, Bmax), max(Amin, Bmin)
        shadow = minmax - maxmin
        if shadow < 0:
            return None
        elif shadow < min_shadow:
            min_shadow = shadow
            norm = u

    # Determina o sentido da normal
    if dot(A._pos, norm) > dot(B._pos, norm):
        norm = -norm

    # Computa o polígono de intersecção e usa o seu centro de massa como ponto
    # de colisão
    try:
        clipped = clip(A.vertices, B.vertices)
    # não houve superposição (talvez por usar normais aproximadas)
    except ValueError:
        return None

    if area(clipped) == 0:
        return None
    col_pt = center_of_mass(clipped)
    return Collision(A, B, pos=col_pt, normal=norm, delta=min_shadow)


###############################################################################
#                 Colisões entre objetos de tipos diferentes
###############################################################################

@get_collision.overload([Circle, AABB])
def circle_aabb(A, B):
    # TODO: implementar direito, está utilizando AABBs
    r = A.cbb_radius
    x, y = A._pos
    Axmin, Axmax = x - r, x + r
    Aymin, Aymax = y - r, y + r

    x, y = B._pos
    dx, dy = B._delta_x, B._delta_y
    Bxmin, Bxmax = x - dx, x + dx
    Bymin, Bymax = y - dy, y + dy

    shadowx = min(Axmax, Bxmax) - max(Axmin, Bxmin)
    shadowy = min(Aymax, Bymax) - max(Aymin, Bymin)
    if shadowx < 0 or shadowy < 0:
        return None

    # Calcula ponto de colisão
    x_col = max(A.xmin, B.xmin) + shadowx / 2.
    y_col = max(A.ymin, B.ymin) + shadowy / 2.
    pos_col = Vec2(x_col, y_col)

    # Define sinal dos vetores normais: colisões tipo PONG
    if shadowx > shadowy:
        n = Vec2(0, (1 if A.ymin < B.ymin else -1))
    else:
        n = Vec2((1 if A.xmin < B.xmin else -1), 0)

    return Collision(A, B, pos=pos_col, normal=n)


@get_collision.overload([AABB, Circle])
def aabb_circle(A, B):
    return circle_aabb(B, A)


@get_collision.overload([Poly, AABB])
def poly_aabb(A, B):
    '''Implementa a colisão entre um polígono arbitrário e uma caixa AABB'''

    return aabb_poly(B, A)


@get_collision.overload([AABB, Poly])
def aabb_poly(A, B):
    '''Implementa a colisão entre um polígono arbitrário e uma caixa AABB'''

    if shadow_x(A, B) < 0 or shadow_y(A, B) < 0:
        return None

    A_poly = Rectangle(bbox=A.bbox)
    col = collision_poly(A_poly, B)
    if col is not None:
        return Collision(A, B, pos=col.pos, normal=col.normal)
    else:
        return None

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from FGAme import *
from FGAme.actions import *

world = World(gravity=800, restitution=0.3, friction=0.8, damping=0.8)
mario = AABB(shape=(50, 100), pos=(325, 150), color='red')
turtle = AABB(shape=(50, 50), pos=(725, 125), color='green')
ground = AABB(shape=(1000, 100), pos=(400, 50), mass='inf')
world.add([mario, ground, turtle])

turtle.damping = 0
turtle.dfriction = 0
turtle.sfriction = 0

action = loop(
    set_velocity(-300, 0) >> delay(0.7)
    >> set_velocity(300, 0) >> delay(0.7)
)
action.start(turtle)


px = 350
jump_vy = 600


@on_key_down('space')
def jump():
    mario.vel += (0, jump_vy)


@on_long_press('left')
def left():
    mario.vel = (-px, mario.vel.y)


@on_long_press('right')
def right():
    mario.vel = (px, mario.vel.y)


@mario.listen('collision')
def on_collision(col):
    A, B = col

world.run()



# -*- coding: utf8 -*-
from random import shuffle, random
from math import sqrt
from FGAme.objects import Poly


def explode(obj, world, energy=0, prob_rec=0.5):
    world.remove(obj)
    N = obj.num_sides
    new_objects = []

    if N == 3:
        sides = [obj.get_side(i).norm() for i in range(3)]
        idx = sides.index(max(sides))
        middle = (obj.vertices[idx] + obj.vertices[(idx + 1) % N]) / 2
        pt0 = obj.vertices[idx]
        pt1 = obj.vertices[(idx + 1) % N]
        pt2 = obj.vertices[(idx + 2) % N]

        new_objects.append(
            Poly([pt0, middle, pt2], color=obj.color, omega=obj.omega, density=obj.density))
        new_objects.append(
            Poly([middle, pt1, pt2], color=obj.color, omega=obj.omega, density=obj.density))

    else:
        # Determina os vértices da triangulação e cria objeto
        for i in range(N):
            pt1 = obj.vertices[i]
            pt2 = obj.vertices[(i + 1) % N]
            pt3 = obj.pos
            new = Poly([pt1,
                        pt2,
                        pt3],
                       color=obj.color,
                       omega=obj.omega,
                       density=obj.density)
            new_objects.append(new)

    # Distribui as energias adicionais aleatoriamente entre os objetos
    N = len(new_objects)
    Z = sum(x ** 2 for x in range(1, N + 1))
    energies = [energy * x ** 2 / Z for x in range(1, N + 1)]
    shuffle(energies)

    # Processa os novos objetos criados
    for i, new in enumerate(new_objects):
        # Calcula a direção radial da explosão e move o objeto
        norm = (new.pos - obj.pos).normalize()
        new.move(norm)

        # Adiciona velocidade de acordo com a velocidade original e com a
        # contribuição de energia
        new.boost(obj.vpoint(new.pos))

        # Adiciona objeto ao mundo
        world.add(new)

        # Aplica a função recursivamente com uma determinada probabilidade
        if random() < prob_rec and new.area > 10:
            explode(new, world, energies[i], 2.0 * prob_rec / N)
        else:
            delta_speed = sqrt(2.0 * energies[i] / new.mass) * norm
            new.boost(delta_speed)



# -*- coding: utf8 -*-

from __future__ import division
from FGAme import *
from random import uniform

import kivy
from kivy.app import App
from kivy.lang import Builder
from kivy.clock import Clock
from kivy.uix.floatlayout import FloatLayout

#=========================================================================
# Cria o mundo
#=========================================================================
rv = lambda: (uniform(-400, 400), uniform(-400, 400))
rp = lambda: (uniform(0, 800), uniform(0, 600))
world = World(background=(0, 0, 0))
world.add_bounds(0, 800, 0, 600)
c1 = Circle(30, vel=rv(), world=world, color='blue')
c2 = Circle(20, vel=rv(), pos=rp(), world=world, color='red')
c3 = Circle(50, vel=rv(), pos=rp(), world=world, color='white')
c4 = Circle(30, vel=rv(), pos=rp(), world=world, color=(0, 255, 0))
c5 = Circle(30, vel=rv(), pos=rp(), world=world, color=(255, 255, 0))
circles = [c1, c2, c3, c4, c5]
world.listen('key-down', 'space', world.stop)
# world.run()


from kivy.graphics import Color, Ellipse, Rectangle

#=========================================================================
# Cria a aplicação kivy
#=========================================================================
kv = '''
<WorldLayout>:
    FloatLayout:
        ToggleButton:
            text: 'Run'
            size_hint_y: 0.1
            size_hint_x: 0.1
            on_press: root.animate()
'''
Builder.load_string(kv)


class WorldLayout(FloatLayout):

    def __init__(self):
        self._instructions = {}
        super(WorldLayout, self).__init__()

        with self.canvas:
            Color(0, 0, 0)
            Rectangle(pos=(0, 0), size=(800, 600))

            for c in circles:
                r = 2 * c.radius
                Color(*c.color.f_rgb)
                kcircle = Ellipse(pos=c.pos_sw, size=(r, r))
                self._instructions[c] = kcircle

    def update_world(self, dt):
        world.update(dt)
        for c, kv in self._instructions.items():
            kv.pos = c.pos_sw

    def animate(self):
        Clock.schedule_interval(self.update_world, 1. / 60)


class RotationApp(App):

    def build(self):
        return WorldLayout()

if __name__ == '__main__':
    RotationApp().run()



from FGAme import *


class GravityWorld(World):

    def init(self):
        # Criamos dois objetos
        A = Circle(20, pos=pos.from_middle(100, 0), vel=(100, 300),
                   color='red')
        B = Circle(20, pos=pos.from_middle(-100, 0), vel=(-100, -300))
        self.A, self.B = A, B
        self.add([A, B])

        # Definimos a força de interação entre ambos
        K = self.K = A.mass
        self.A.force = lambda t: -K * (A.pos - B.pos)
        self.B.force = lambda t: -K * (B.pos - A.pos)

        # Redefinimos a constante de amortecimento
        self.damping = 0.5

        # Definimos uma margem de 10px de espessura
        self.add_bounds(width=10)

    @listen('key-down', 'space')
    def toggle(self):
        self.toggle_pause()

    @listen('long-press', 'right')
    def move_right(self):
        self.A.move(5, 0)

    @listen('long-press', 'left')
    def move_left(self):
        self.A.move(-5, 0)

    @listen('long-press', 'up')
    def move_up(self):
        self.A.move(0, 5)

    @listen('long-press', 'down')
    def move_down(self):
        self.A.move(0, -5)

    @listen('mouse-button-down', 'left')
    def add_circle(self, pos):
        self.pause()
        self.circle = Circle(20, pos=pos, color='random')
        self.line = draw.Segment(pos, pos)
        self.add([self.circle, self.line])

    @listen('mouse-long-press', 'left')
    def set_circle_velocity(self, pos):
        self.line.end = pos

    @listen('mouse-button-up', 'left')
    def launch_circle(self, pos):
        self.unpause()
        self.remove(self.line)
        self.circle.boost(4 * self.line.direction)

if __name__ == '__main__':
    world = GravityWorld()
    world.run()



from functools import partial


def pyname(value):
    '''Retorna versão do nome aceitável como nome de variável Python
    (e.g.: troca '-' por '_') e outras substituições.'''

    return value.replace('-', '_')


def wrap_handler(func, args, kwds, pre_args=0):
    '''Fixa os argumentos posicionais e keywords na função.
    Retorna uma função que sempre exige pre_args argumentos posicionais'''

    func_out = func

    if args and kwds:
        if pre_args == 0:
            func_out = partial(func, *args, **kwds)
        elif pre_args == 1:
            def func_out(x):
                return func(x, *args, **kwds)
        elif pre_args == 2:
            def func_out(x, y):
                return func(x, y, *args, **kwds)
        else:
            def func_out(*f_args):
                f_args += args
                return func(*f_args, **kwds)

    elif args:
        if pre_args == 0:
            func_out = partial(func, *args)
        elif pre_args == 1:
            def func_out(x):
                return func(x, *args)
        elif pre_args == 2:
            def func_out(x, y):
                return func(x, y, *args)
        else:
            def func_out(*f_args):
                f_args += args
                return func(*f_args)

    elif kwds:
        func_out = partial(func, **kwds)

    return func_out


def listen(name, *args, **kwds):
    '''Decorador que registra uma função como sendo um callback de um
    determinado sinal'''

    def decorator(func):
        try:
            L = func._listen_args
        except AttributeError:
            L = func._listen_args = []
        finally:
            L.append((name, args, kwds))
        return func

    return decorator


def signal(name, *filters, **kwds):
    '''Define um sinal.

    Possui a assinatura signal(nome, [arg1, ...,[ num_args]]), onde arg_i são
    strings com os nomes dos argumentos e o valor opcional num_args é um número
    representando o número de argumentos para o handler padrão.

    O tipo de sinal é escolhido automaticamente a partir dos argumentos:

    >>> signal('foo', num_args=1)
    Signal('foo', 1)

    >>> signal('foo', 'bar')
    FilteredSignal('foo', 0, filter_names=['bar'])

    >>> signal('foo', delegate_to='_foo_attr')
    DelegateSignal('foo', 0)
    '''

    from FGAme.events.signals import Signal, DelegateSignal, FilteredSignal

    # Extrai argumentos
    num_filters = len(filters)
    num_args = kwds.pop('num_args', 0)
    delegate_to = kwds.pop('delegate_to', None)
    if kwds:
        raise TypeError('invalid argument: %r' % kwds.popitem()[0])

    if delegate_to:
        if num_filters == 0:
            return DelegateSignal(name, delegate_to=delegate_to)
        else:
            return DelegateSignal(name,
                                  delegate_to=delegate_to,
                                  num_filters=num_filters,
                                  filter_names=filters)
    elif num_filters == 0:
        return Signal(name, num_args=num_args)
    elif num_filters == 1:
        return FilteredSignal(name, num_args=num_args, filter_names=filters)
    else:
        raise NotImplementedError('filtered signals with more than 1 argument')



#-*- coding: utf8 -*-
'''
Benchmarks para FGAme
'''

import FGAme
import time
BASE_SIZE = 1000
BENCH_FUNCTIONS = []
try:
    range = xrange  # @ReservedAssignment @UndefinedVariable  
except NameError:
    pass


def bench(func):
    '''Decorador para ser aplicado nas funções de teste de benchmarks'''

    BENCH_FUNCTIONS.append(func)
    return func


def all(size=None):
    '''Executa todos os benchmarks no módulo'''

    for func in BENCH_FUNCTIONS:
        t0 = time.time()
        text = func.__doc__
        func_name = func.__name__
        func()
        t1 = time.time()
        print('%s: (%s) %.3f sec' % (func_name, text, t1 - t0))

#===============================================================================
# Vetores
#===============================================================================


@bench
def vector():
    '''Operações aritiméticas com vetores'''

    v = FGAme.Vec2(1.5, 1.5)
    for _ in range(BASE_SIZE * 1000):
        v = 2 * v - v


@bench
def ivector():
    '''Operações aritiméticas inplace com vetores'''

    v = FGAme.Vec2(1.5, 1.5)
    v2 = FGAme.mVec2(1.5, 1.5)
    for _ in range(BASE_SIZE * 1000):
        v2 += v


@bench
def vtuple():
    '''Operações aritiméticas vetores e tuplas'''

    v = FGAme.Vec2(1.5, 1.5)
    v2 = FGAme.mVec2(1.5, 1.5)
    for _ in range(BASE_SIZE * 1000):
        v2 += (1, 1)
        v += (1, 1)

if __name__ == '__main__':
    all()


# -*- coding: utf8 -*-
'''
Este exemplo ilustra a simulação de um gás de caixas alinhadas aos eixos
'''
from FGAme import *
from random import uniform, choice

# Constantes da simulação
SPEED = 300
RADIUS = 15
N = 50

# Inicializa o mundo
world = World()
world.add_bounds(width=10)

# Preenche o mundo
for _ in range(N):
    pos = Vec2(uniform(30, 770), uniform(30, 570))
    vel = Vec2(uniform(-SPEED, SPEED), uniform(-SPEED, SPEED))
    aabb = Circle(RADIUS, vel=vel, pos=pos, color=(200, 0, 0))
    world.add(aabb)

# Inicia a simulação
world.register_energy_tracker()
world.run()



#-*-coding: utf8 -*-

from FGAme import draw, conf
from time import sleep

# Cria a árvore de renderização com vários círculos
p0 = (400, 300)
tree = draw.RenderTree()
tree.add(draw.Circle(150, p0, color='red'))
tree.add(draw.Circle(120, p0, color='white'))
tree.add(draw.Circle(100, p0, color='red'))
tree.add(draw.Circle(70, p0, color='white'))
tree.add(draw.Circle(50, p0, color='red', line_color='black', line_width=10))

# Cria um objeto Canvas com a geometria da tela
canvas = conf.init_screen(800, 600)
canvas.show()

# Loop principal
for i in range(50):
    canvas.clear_background('white')
    canvas.draw(tree)
    canvas.flip()
    sleep(0.01)


'''

Especificação de tilesets
=========================

...


::

    conf:
        origin: 0, 0
        shape: 50, 50
        background-color: #99f

    tile "brick":
        char: "x"
        color: red
        image: tiles/brick.png

    tile "coin":
        char: "o"
        shape: circle
        color: #ff0

    tile "spike":
        char: "i"
        image: tiles/spike.png

    data "main":
        |             oo
        |        oo     xxxxxxxxxxxxxxxxxx
        |           xxx
        |       xxx
        |                                        ii
        |xxxxxxxxxxxxxxxxxxxxxxxxxxxxx    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxiiiixxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

'''



from __future__ import division
from collections import UserList
from FGAme import *


class Integrator(UserList):

    '''Classe se comporta como uma lista de objetos cuja dinâmica é resolvida
    separadamente do loop principal da FGAme.
    Objetos podem ser adicionados a um grupo do tipo Integrator caso o
    algoritmo de Euler semi-implícito utilizado pela FGAme não possua a
    acurácia necessária para resolução de forças (ex.: sistemas
    autogravitantes)
    '''

    def __init__(self):
        self._data = []

    def add(self, obj):
        '''Adiciona um objeto à lista de objetos'''

        self.append(obj)

    @property
    def data(self):
        return self._data

    def update(self, dt):
        '''Atualiza as posições e velocidades'''


class GravityIntegrator(Integrator):

    def __init__(self, G=1, framerate=30):
        super(GravityIntegrator, self).__init__()
        self.G = G
        self.dt = 1 / framerate

    def update_accelerations(self):
        '''Retorna a lista de acelerações para cada objeto'''

        for i, current in enumerate(self):
            current.accel_prev = current.accel

            for other in self:
                if current != other:
                    #d  = r(b) - r(a)
                    distance = other.pos - current.pos
                    # a(a) = m(b) * d/|d|^3
                    current.accel = other.mass * \
                        distance / distance.norm() ** 3
            # a <= a*G
            current.accel *= self.G
            self[i] = current

    def update(self, iterations=1):
        '''Utiliza o método de Velocity Verlet para atualizar as posições e
        velocidades'''

        dt = self.dt / iterations
        for _ in range(iterations):
            self.update_accelerations()

            for i, p in enumerate(self):
                # r(k+1) = r(k) + dt*v(k) + a(k)*(dt**2) /2
                p.pos += dt * p.vel + p.accel_prev * (dt ** 2) / 2
                # v(k+1) = v(k) + dt * ( a(k)+a(k-1) )/2
                p.vel += + dt * (p.accel + p.accel_prev) / 2

                self[i] = p


#### TESTE UNITÁRIO ######
#'''
# Este exemplo simula um sitema solar com dois corpos, que possuem
# massas, posições e velociodades adequados para manter uma orbita.
class Universe(World):

    def __init__(self):
        super(Universe, self).__init__(background=(35, 0, 100))

        sun = Circle(radius=5, color=(255, 215, 0), mass=1e3,
                     pos=(400, 300), vel=(0, 0))
        earth = Circle(radius=3, color=(0, 200, 255), mass=1,
                       pos=(400 + 80, 300), vel=(0, (1e3 / 200) ** 0.5))
        self.orbit = GravityIntegrator()
        self.orbit.add(sun)
        self.orbit.add(earth)
        self.add((sun, earth))

    @listen('frame-enter')
    def resolve(self):
        for i in range(10):
            self.orbit.update()
if __name__ == '__main__':
    universe = Universe()
    universe.run()
#'''



# -*- coding: utf8 -*-
from FGAme import *

# Cria mundo e objetos
world = World()
world.add_bounds(width=10)
c1 = Circle(50, pos=(50, 300), vel=(800, 0), color='random')
c2 = Circle(30, pos=(420, 320), color='random')
c3 = Circle(30, pos=(420, 280), color='random')

# Insere objetos e inicia rastreio de energia
world.register_energy_tracker()
world.add([c1, c2, c3])
world.run()



# -*- coding: utf8 -*-
'''
======
Forças
======

Forças externas podem ser inseridas manualmente nos objetos físicos da FGAme
através do atributo obj.force das instâncias de `Body()` ou criadas através de
um objeto da classe Force().

No caso mais simples, podemos simplesmente atribuir uma função que retorna o
vetor de força como função do tempo:

>>> from FGAme import Circle
>>> obj = Circle(5)
>>> obj.force = lambda t: (4, 3) - obj.pos

Neste caso, podemos simplesmente calcular a força chamando

>>> obj.force(0)
Vec(4, 3)

O attributo obj.force aceita várias regras de composição. A mais útil delas,
de soma de forças, permite aplicar várias forças externas a um mesmo objeto.

>>> obj.force += lambda t: (10, 10)

Neste caso, a composição de forças corresponde à adição dos valores de cada
componente

>>> obj.force(0)
Vec(14, 13)


Objetos do tipo Force()
=======================

Caso se trate de uma força comum, é mais conveniente utilizar uma classe que
automatiza a criação e configuração da força do objeto. Existem várias classes
pré-definidas, implementando forças como molas, gravidade, etc.


Forças simples
--------------

Estas são as forças que atuam em uma única partícula. Todas elas herdam da
classe `Single`. Existem ainda as classes `Conservative` e `Dissipative` que
implementam forças conservativas e aquelas que dependem apenas da velocidade.

Para criar uma força de atração gravitacional para o centro da tela, por
exemplo, basta criar o objeto do tipo Gravity() e atribuir ao atributo
obj.force:

>>> obj.force = Gravity(obj, G=1e5, pos=(400, 300))

Como conveniência, podemos também usar a função set_force_gravity():

>>> set_force_gravity(obj, G=1e5, pos=(400, 300))          # doctest: +ELLIPSIS
<...>

.. autoclass::Gravity
.. autoclass::Spring
.. autoclass::SpringTensor
.. autoclass::Viscous
.. autoclass::Drag
.. autoclass::Friction


Forças de pares
---------------

São forças que atuam em um par de partículas respeitando a Terceira Lei de
Newton (lei da ação e reação). Elas podem derivar de `Pair` ou
`ConservativePair`.

.. autoclass::GravityPair
.. autoclass::SpringPair
.. autoclass::SpringTensorPair


Sistemas de forças
------------------

São forças aplicáveis a qualquer número arbitrariamente grande de partículas.
Podem corresponder a sistemas auto-gravitantes, ou forças de molas, etc. Herdam
de `Pool` ou de `ConservativePool`

.. autoclass::GravityPool
.. autoclass::SpringPool
'''

from FGAme.mathtools import Vec2, null2D, asvector

EMPTY_FORCE = lambda t: null2D
NO_UPDATE = object()


class ForceProperty(object):

    '''Implementa o atributo obj.force dos objetos da classe Body.

    ForceProperty facilita a composição de forças num mesmo objeto e apresenta
    um sintaxe conveniente para especificar as forças atuantes em um objeto.

    Exemplo
    -------

    Considere uma classe com uma propriedade do tipo ForceProperty()

    >>> class HasForce(object):
    ...     force = ForceProperty()

    Criamos um objeto desta classe e atribuimos uma força a ele.

    >>> obj = HasForce()
    >>> obj.force = lambda x: (0, x**2)
    >>> obj.force(2)
    Vec(0, 4)

    Também podemos definir a força utilizando um decorador se for mais
    conveniente

    >>> @obj.force.define
    ... def force(x):
    ...     return (0, x**3)
    >>> obj.force(2)
    Vec(0, 8)

    O recurso mais importante, no entanto, é a composição de forças. Podemos
    definir uma composição de forças simplesmente adicionando ou subtraindo
    forças ao atributo original

    >>> obj.force += lambda x: (0, x**2)
    >>> obj.force(2) # => 2**2 + 2**3
    Vec(0, 12)

    Outras operações algébricas também estão definidas

    >>> obj.force *= 2
    >>> obj.force(2)
    Vec(0, 24)

    Os decoradores também estão disponíveis para as operações de adição e
    subtração de forças.

    >>> @obj.force.add
    ... def force(x):
    ...     return (0, x**4)
    >>> obj.force(2)
    Vec(0, 40)

    Notas
    -----

    A função que calcula a força será sempre gravada no atributo obj._force. O
    acesso via este atributo é mais rápido e pode ser utilizado internamente
    para os implementadores da classe.

    Um segundo atributo _force_ctrl também é criado para guardar o objeto que
    controla e coordena a composição de forças e os outros recursos da
    ForceProperty.

    O trabalho pesado é feito pela class ForceCtrl. Todos os acessos à
    obj.force do exemplo anterior simplesmente retornam o valor do atributo
    obj._force_ctl, que é a instância de ForceCtrl que realmente faz a
    coordenação das forças

    >>> obj._force_ctrl                                    # doctest: +ELLIPSIS
    <...ForceCtrl object at 0x...>
    >>> obj.force                                          # doctest: +ELLIPSIS
    <...ForceCtrl object at 0x...>

    Lembramos que obj._force é uma função simples que executa mais rapidamente
    que obj.force pois não precisa passar pelo mecanismo que implementa as
    propriedades no Python.

    >>> obj._force # doctest: +ELLIPSIS
    <function ... at 0x...>

    Caso não possua nenhuma força registrada no objeto, é garantido que o
    atributo obj._force possui a identidade da função EMPTY_FORCE, que recebe
    um argumento temporal e retorna sempre o vetor nulo.
    '''

    def __get__(self, obj, cls):
        if obj is None:
            return self
        try:
            return obj._force_ctrl
        except AttributeError:
            obj._force_ctrl = ctrl = ForceCtrl(obj)
            obj._force = EMPTY_FORCE
            return ctrl

    def __set__(self, obj, func):
        # matemática inplace acaba gravando valores espúrios na força.
        # obj.x.__iadd__() é o mesmo que obj.x = y.__iadd__(), onde y é o valor
        # retornado por obj.x. Deste modo, o retorno de __iadd__ acaba sendo
        # atribuido novamente para obj.x. A solução aqui foi fazer __iadd__
        # retornar um objeto especial NO_UPDATE e checar explicitamente se
        # este está sendo atribuído à propriedade.
        if func is not NO_UPDATE:
            ctrl = self.__get__(obj, type(obj))
            ctrl.define(func)


class ForceCtrl(object):

    '''
    Controla as operações com forças em um objeto. Pode ser chamado com um
    argumento numérico para calcular a força. Também aceita o idioma de
    composição de forças descrito na classe ForceProperty
    '''

    def __init__(self, obj):
        self._funcs = []
        self._obj = obj
        self._fast = EMPTY_FORCE

    def define(self, force):
        '''Limpa todas as forças anteriores e redefine o objeto de força.

        Pode ser chamado como um decorador.'''

        self.clear()
        self.add(force)
        return force

    def add(self, force):
        '''Adiciona uma nova força à interação com a partícula.

        O resultado é uma força cujo resultado é a soma de force(t) com as
        forças já definidas anteriormente.
        '''

        self._funcs.append((None, force))
        self._update_fast()
        return force

    def mul(self, factor):
        '''Multiplica todas as forças já definidas por uma constante
        multiplicativa. A constante deve ser um número.'''

        assert not callable(factor)
        L = self._funcs
        for i, (k, f) in enumerate(L):
            k = factor if k is None else factor * k
            L[i] = (k, f)
        self._update_fast()

    def clear(self):
        '''Limpa todas as forças que atuam no objeto'''

        self._funcs = []
        self._update_fast()

    # Funções privadas -------------------------------------------------------
    def _update_fast(self):
        '''Atualiza a função que calcula a força de acordo com os tipos de
        transformações presentes até o momento.

        Atualiza o método de acesso rápido do objeto em questão.'''
        if self._funcs:
            # Monta hierarquia de funções rápidas
            adds = [f for (k, f) in self._funcs if k is None]
            muls = [(k, f) for (k, f) in self._funcs if k is not None]
            fmuls = [(k, f) for (k, f) in muls if callable(k)]
            nmuls = [(k, f) for (k, f) in muls if not callable(k)]

            #
            # Implementa versões específicas, da mais especializada para a
            # menos especializada.
            #

            # Começamos no caso onde só existem forças aditivas
            if not fmuls and not nmuls:
                fast_func = self._make_fast_adds(adds)

            # Agora o caso onde todas as multiplicações são numéricas. Criamos
            # uma lista de tuplas (k, f) com a constante multiplicativa e
            # com a função
            elif not fmuls:
                nmuls += [(1, f) for f in adds]
                fast_func = self._make_fast_nmuls(nmuls)

            # Agora existem algumas multiplicações por funções
            else:
                nmuls += [(1, f) for f in adds]
                fmuls += [(lambda x: 1, f) for f in adds]
                fast_func = self._make_fast_fmuls(fmuls)

            # Salva a função rápida
            self._obj._force = self._fast = fast_func

        # Não existe nenhuma função registrada: atribui o atributo de acesso
        # rápido a None
        else:
            self._fast = EMPTY_FORCE
            self._obj._force = EMPTY_FORCE

    def _make_fast_adds(self, adds):
        '''Cria função que retorna a força como a soma de todas as forças na
        lista adds'''

        adds = tuple(adds)
        assert None not in adds, adds

        def fast_func(t):
            F = null2D
            for func in adds:
                F += func(t)
            return F
        return fast_func

    def _make_fast_nmuls(self, nmuls):
        '''Cria função que retorna a força como a soma do produto de cada
        constante k com a força f presente como as tuplas (k, f) em nmuls'''

        nmuls = tuple(nmuls)

        def fast_func(t):
            F = null2D
            for k, func in nmuls:
                Fi = asvector(func(t))
                try:
                    F += Fi * k
                except TypeError:
                    x, y = Fi
                    F += Vec2(k * x, k * y)
            return F
        return fast_func

    def _make_fast_fmuls(self, fmuls):
        '''Cria função que retorna a força como a soma do produto de cada
        constante k(t) com a força f presente como as tuplas (k, f) em fmuls'''

        fmuls = tuple(fmuls)

        def fast_func(t):
            F = null2D
            for scalar_func, func in fmuls:
                Fi = func(t)
                k = scalar_func(t)
                try:
                    F += Fi * k
                except TypeError:
                    x, y = Fi
                    F += Vec2(k * x, k * y)
            return F
        return fast_func

    # Métodos mágicos --------------------------------------------------------
    def __call__(self, t):
        return self._fast(t)

    def __iadd__(self, other):
        self.add(other)
        return NO_UPDATE

    def __imul__(self, other):
        self.mul(other)
        return NO_UPDATE


class Force(object):

    '''Implementa uma força que atua em um conjunto de objetos.'''

    @classmethod
    def setting_force(cls, *args, **kwds):
        '''Inicializa a força e já registra no atributo obj.force de todos os
        objetos envolvidos'''

        if 'register' in kwds:
            raise TypeError('invalid argument: register')

        kwds['register'] = True
        return cls(*args, **kwds)


###############################################################################
#         Implementações de forças específicas -- forças de 1 objeto
###############################################################################
def set_force_single(obj, func, mode):
    '''Define uma força simples sobre a partícula `obj`.

    Parameters
    ----------

    obj : Body
        Objeto em que a força atua
    force : callable
        Uma função que retorna o vetor com a força resultante. A assinatura da
        função depende do argumento `mode`.
    mode : str
        Existem quatro possibilidades. Cada um assume uma assinatura
        específica.
            'none':
                force() -> F
            'time':
                force(t) -> F
            'object':
                force(obj) -> F
            'position':
                force(obj.pos) -> F
            'velocity':
                force(obj.vel) -> F
            'all':
                force(obj.pos, obj.vel, t) -> F
        O valor padrão é 'time', que corresponde à convenção empregada pelo
        atributo obj.force dos objetos físicos.

    See also
    --------

    `Single`
    '''

    return Single.setting_force(obj, func, mode)


class Single(Force):

    '''Força simples aplicada a um objeto.

    Ver `set_force_single()` para descrição dos argumentos.'''

    def __init__(self, obj, force, mode='time', register=False):
        self._obj = obj
        self._force = force
        self._mode = mode

        # Converte force para a assinatura force(t) -> F
        if mode == 'time':
            self._worker = force
        elif mode == 'none':
            self._worker = lambda t: force()
        elif mode == 'object':
            self._worker = lambda t: force(obj)
        elif mode == 'all':
            self._worker = lambda t: force(obj, t)
        elif mode == 'position':
            self._worker = lambda t: force(obj.pos)
        elif mode == 'velocity':
            self._worker = lambda t: force(obj.vel)
        else:
            raise ValueError('invalid mode: %r' % mode)

        if register:
            self._obj.force.add(self._worker)

    def __call__(self, t):
        return self._worker(t)

    obj = property(lambda self: self._obj)
    force = property(lambda self: self._force)
    mode = property(lambda self: self._mode)


def set_force_viscous(obj, gamma):
    ''''''
    # TODO: documente-me!
    return Viscous.setting_force(obj, gamma)


class Viscous(Single):

    def __init__(self, obj, gamma=0, register=False):
        self.gamma = gamma
        self._viscous_force = force = lambda v: -gamma * obj.vel
        super(Viscous, self).__init__(obj, force, 'velocity', register)


def set_force_conservative(obj, force, potential):
    '''Força conservativa que atua em um único objeto.

    Parameters
    ----------
    obj : Body
        Objeto em que a força atua
    force : callable
        Função do vetor posição que retorna a força resultante.
    potential : callable
        Função do vetor posição que retorna a energia potencial.

    See also
    --------

    `Conservative`

    '''
    return Conservative.setting_force(obj, force, potential)


class Conservative(Single):

    '''Força conservativa que atua em um único objeto.

    Veja `set_force_conservative()` para descrição dos argumentos.'''

    def __init__(self, obj, force, potential, register=False):
        super(Conservative, self).__init__(
            obj, force, 'position', register=register)
        self._potential = potential

    def energyT(self):
        '''Energia total do par de partículas'''

        return self.obj.energyK() + self.energyU()

    def energyK(self):
        '''Energia cinética do par de partículas'''

        return self.obj.energyK()

    def energyU(self):
        '''Energia potencial do par de partículas'''

        return self.U(self.obj.pos)

    potential = property(lambda self: self._potential)


def set_force_gravity(obj, G, M=None, epsilon=0, pos=(0, 0)):
    '''Força de gravitacional produzida por um objeto de massa "M" fixo na
    posição "pos".

    A expressão para o potencial gravitacional é

        U(r) = - G M m / (|r - pos| + epsilon),

    onde m é a massa de obj e r a sua posição. A expressão anterior resulta na
    força

        F(r) = - G M m / (|r - pos| + epsilon)**2 * u,

    onde u = (r - pos) / |r - pos| é o vetor unitário na direção da linha que
    separa os centros.

    Parameters
    ----------
    obj : Body
        Objeto em que a força atua
    G : float
        Constante gravitacional. Como a FGAme não utiliza unidades físicas, não
        faz sentido atribuir um valor padrão para G. Deve ser calibrado em cada
        caso para adequar à dinâmica do jogo.
    M : float
        Massa da fonte gravitacional. Não precisa ser fornecido, já que o que
        importa é o produto entre M e G.
    epsilon : float
        Parâmetro de suavização da força da gravidade. Impede que ela exploda
        quando a distância tende a zero. Para distâncias menores que epsilon,
        a força gravitacional começa a se assemelhar a uma força de mola.
    pos : Vec2
        Vetor com a posição do centro da fonte de gravidade.

    See also
    --------

    `Gravity`
    '''

    return Gravity.setting_force(obj, G=G, M=M, epsilon=epsilon, pos=pos)


class Gravity(Conservative):

    '''Força de gravitacional produzida por um objeto de massa "M" fixo na
    posição "pos".

    Veja `set_force_gravity()` para descrição dos argumentos.'''

    def __init__(self, obj, G, M=None, epsilon=0, pos=(0, 0), register=False):
        if M is None:
            M = obj.mass
        self._M = M = float(M)
        self._epsilon = epsilon = float(epsilon)
        self._G = G = float(G)
        self._pos = pos = Vec2(*pos)

        def F(R):
            R = pos - R
            r = R.norm()
            r3 = (r + epsilon) ** 2 * r
            R *= G * obj.mass * M / r3
            return R

        def U(R):
            R = (R - pos).norm()
            return -self._G * obj.mass * M / (R + epsilon)

        super(Gravity, self).__init__(obj, F, U, register=register)

    G = property(lambda self: self._G)
    M = property(lambda self: self._M)
    epsilon = property(lambda self: self._epsilon)
    pos = property(lambda self: self._pos)


def set_force_spring(obj, k, gamma=0, pos=(0, 0)):
    # TODO: documentação
    return Spring.setting_force(obj, k, gamma, pos)


class Spring(Conservative):
    # TODO, implementar mola regular isotrópica e amortecida

    def __init__(self, k, gamma=0, pos=(0, 0)):
        raise NotImplementedError


def set_force_spring_tensor(obj, k, pos=(0, 0)):
    # TODO: documentação
    return SpringTensor.setting_force(obj, k, pos)


class SpringTensor(Conservative):

    '''Implementa uma força devido a uma mola com constante elástica k fixa na
    posição r0. Consulte a documentação de SpringTensorPair() para verificar
    como configurar a anisotropia e parâmetros da mola.'''

    def __init__(self, obj, k, pos=(0, 0), register=False):
        self._k = k
        kxy = 0
        x0, y0 = self._pos = Vec2(*pos)

        try:  # Caso isotrópico
            kx = ky = k = self._k = float(k)
        except TypeError:  # Caso anisotrópico
            k = self._k = tuple(map(float, k))
            if len(k) == 2:
                kx, ky = k
            else:
                kx, ky, kxy = k

        # Define forças e potenciais
        def F(R):
            Dx = x0 - R.x
            Dy = y0 - R.y
            return Vec2(kx * Dx + kxy * Dy, ky * Dy + kxy * Dx)

        def U(R):
            Dx = x0 - R.x
            Dy = y0 - R.y
            return (kx * Dx ** 2 + ky * Dy ** 2 + 2 * kxy * Dx * Dy) / 2.0

        super(SpringTensor, self).__init__(obj, F, U, register=register)

    k = property(lambda self: self._k)
    r0 = property(lambda self: self._pos)


###############################################################################
#        Implementações de forças específicas -- forças de 2 objetos
###############################################################################
def set_force_pair(A, B, func, mode='time'):
    '''
    Força que opera em um par de partículas respeitando a lei de ação e reação.

    Parameters
    ----------

    A, B : Object instances
        Objetos em que a força atua
    force : callable
        Uma função que retorna o vetor com a força resultante. A assinatura da
        função depende do último argumento `mode`. Por convenção é a força que
        A produz em B (ou força sofrida por B devido à A).
    mode : str
        Existem cinco possibilidades. Cada um assume uma assinatura específica.
            'none':
                force() -> F
            'time':
                force(t) -> F
            'object':
                force(A, B) -> F
            'position':
                force(A.pos, B.pos) -> F
            'all':
                force(A.pos, B.pos, t) -> F
        O valor padrão é 'time', que corresponde à convenção empregada pelo
        atributo obj.force dos objetos físicos.

    See also
    --------

    `Pair`
    '''
    return Pair.setting_force(A, B, func, mode, True)


class Pair(Force):

    '''
    Força que opera em um par de partículas respeitando a lei de ação e reação.

    Veja `set_force_pair()` para descrição dos argumentos.'''

    def __init__(self, A, B, func, mode='time', register=False):
        self._A = A
        self._B = B
        self._force = func
        self._mode = mode
        self._cacheF = None

        # Converte force para a assinatura force(t) -> F
        if mode == 'time':
            self._worker = func
        elif mode == 'none':
            self._worker = lambda t: func()
        elif mode == 'object':
            self._worker = lambda t: func(A, B)
        elif mode == 'all':
            self._worker = lambda t: func(A.pos, B.pos, t)
        elif mode == 'position':
            self._worker = lambda t: func(A.pos, B.pos)
        else:
            raise ValueError('invalid mode: : %r' % mode)

        # Registra as forças
        if register:
            self._A.force.add(self.force_A)
            self._B.force.add(self.force_B)

    # Atributos somente para leitura
    A = property(lambda self: self._A)
    B = property(lambda self: self._B)
    force = property(lambda self: self._force)
    mode = property(lambda self: self._mode)

    def force_A(self, t):
        '''Função que calcula a força sobre o objeto A no instante t'''

        if self._cacheF is None:
            res = self._worker(t)
            self._cacheF = -res
            return res
        else:
            res = self._cacheF
            self._cacheF = None
            return res

    def force_B(self, t):
        '''Função que calcula a força sobre o objeto B no instante t'''

        if self._cacheF is None:
            res = self._worker(t)
            self._cacheF = res
            return -res
        else:
            res = self._cacheF
            self._cacheF = None
            return res

    def accel_A(self, t):
        '''Função que calcula a aceleração sobre o objeto A no instante t'''

        return self.force_A(t) / self._A.mass

    def accel_B(self, t):
        '''Função que calcula a aceleração sobre o objeto B no instante t'''

        return self.force_A(t) / self._A.mass

    def __iter__(self):
        yield self.force_A
        yield self.force_B

    def forces(self):
        '''Itera sobre as funções [force_A, force_B]'''

        yield self.force_A
        yield self.force_B

    def accels(self):
        '''Itera sobre as acelerações [accel_A, accel_B]'''

        yield self.accel_A
        yield self.accel_B


def set_force_conservative_pair(A, B, force, U):
    # TODO: documentação
    return ConservativePair.setting_force(A, B, force, U)


class ConservativePair(Pair):

    '''
    Uma força que opera em um par de partículas e possui uma energia potencial
    associada.

    Esta classe é iniciada a partir de uma função force(rA, rB) e de um
    potencial U(rA, rB).
    '''

    def __init__(self, A, B, force, U, register=False):
        super(ConservativePair, self).__init__(
            A, B, force, 'position', register=register)
        self._potential = U

    def energyT(self):
        '''Energia total do par de partículas'''

        return self.A.energyK() + self.B.energyK() + self.energyU()

    def energyK(self):
        '''Energia cinética do par de partículas'''

        return self.A.energyK + self.B.energyK

    def energyU(self):
        '''Energia potencial do par de partículas'''

        return self.U(self.A.pos, self.B.pos)

    U = property(lambda self: self._potential)


class SpringPair(Conservative):
    # TODO, implementar spring regular isotrópica e amortecida com uma
    # distância fixa de atuação.
    pass


def set_force_spring_tensor_pair(A, B, k):
    '''
    Liga as duas partículas por uma força de Hooke com uma constante de mola
    anisotrópica.

    O potencial é dado por::

        U = kx * dx**2 / 2 + ky * dy**2 /2 + kxy * dx * dy

    Onde dx e dy são as diferenças de posições nas direções x e y
    respectivamente. Este potencial resulta em uma força::

        F = -kx * dx - ky * dy -kxy * (dx + dy)

    Os termos dx e dy são as componentes do vetor ``B.pos - A.pos``.

    Parameters
    ----------

    A, B : Body
        Objetos que participam da interação de mola.
    k : float ou tupla
        Especifica a constante de mola como sendo isotrópica (se k é for um
        número) ou através de uma tuplas (kx, ky) ou (kx, ky, kxy).

    See Also
    --------

    `SpringTensorPair`
    '''
    return SpringTensorPair.setting_force(A, B, k)


class SpringTensorPair(ConservativePair):

    '''
    Liga as duas partículas por uma força de Hooke com uma constante de mola
    anisotrópica.

    Veja `set_force_string_tensor_pair()` para descrição dos argumentos.'''

    def __init__(self, A, B, k, register=False):
        kxy = 0
        try:  # Caso isotrópico
            kx = ky = k = self._k = float(k)
        except TypeError:  # Caso anisotrópico
            k = self._k = tuple(map(float, k))
            if len(k) == 2:
                kx, ky = k
            else:
                kx, ky, kxy = k

        # Define forças e potenciais
        def F(rA, rB):
            dx, dy = rB - rA
            return Vec2(kx * dx + kxy * dy, +ky * dy + kxy * dx)

        def U(rA, rB):
            dx, dy = rB - rA
            return (kx * dx ** 2 + ky * dy ** 2 + 2 * kxy * dx * dy) / 2.0

        super(SpringTensorPair, self).__init__(A, B, F, U, register=register)

    k = property(lambda self: self._k)
    delta = property(lambda self: self._delta)


def set_force_gravity_pair(A, B, G, epsilon=0):
    # TODO: documentação
    return GravityPair.setting_force(A, B, G, epsilon)


class GravityPair(ConservativePair):

    '''
    Implementa a força de gravidade "amaciada" com um parâmetro epsilon.

    A energia potencial desta força é dada por

        U(rA, rB) = -G mA mB / (|rB - rA| + e)

    O parâmetro de amaciamento pode ser necessário para melhorar a estabilidade
    numérica caso as duas posições rA e rB se aproximem muito. O valor padrão
    é zero e o usuário deve ponderar sobre valores adequados caso a caso. Na
    prática epsilon limita o valor máximo que a força gravitacional pode
    assumir.
    '''

    def __init__(self, A, B, G, epsilon=0, register=False):
        self._G = G = float(G)
        self._epsilon = epsilon = float(epsilon)

        def F(Ra, Rb):
            R = Rb - Ra
            r = R.norm()
            r3 = (r + epsilon) ** 2 * r
            R *= G * A.mass * B.mass / r3
            return R

        def U(Ra, Rb):
            R = (Ra - Rb).norm()
            return -self._G * B.mass * A.mass / (R + self._epsilon)

        super(GravityPair, self).__init__(A, B, F, U, register)

    G = property(lambda self: self._G)
    epsilon = property(lambda self: self._epsilon)


###############################################################################
# Implementações de forças específicas -- forças aplicadas a grupos de objetos
###############################################################################
class Pool(Force):

    '''Força que atua em um grupo arbitrariamente grande de objetos'''

    def __init__(self, objects, register=False):
        self.objects = list(objects)


class ConservativePool(Pool):

    '''Sistema de partículas com força conservativa'''


class GravityPool(ConservativePool):

    '''Força gravitacional entre um grupo arbitrariamente grande de objetos.'''

    def get_force(self, obj, mutable=True):
        pass

    def get_all_forces(self, mutable=True):
        return [self.get_force(obj, mutable) for obj in self.objects]
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
if __name__ == '__main__':
    class HasForce(object):
        force = ForceProperty()

    obj = HasForce()
    obj.force = lambda t: (2, 4 - t)
    obj.force(2)
    obj.force *= 2
    obj.force(2)

    import doctest
    doctest.testmod()

    from math import cos
    from FGAme import *
    from FGAme.physics import set_force_gravity_pair

    world = World(restitution=0)
    c1 = Circle(10, density=2, pos=(300, 200), vel=(100, -100), color='red')
    c2 = Circle(10, pos=(500, 400), vel=(-100, 100))
    c3 = Circle(10, pos=(200, 300), vel=(-100, 100))

    #k = 1000
    #c1.force = lambda t: -k * (c1.pos - c2.pos)
    #c2.force = lambda t: -k * (c2.pos - c1.pos)

    G = 1.1e5
    e = 200
    set_force_gravity_pair(c1, c2, G=G, epsilon=e)
    set_force_gravity_pair(c1, c3, G=G, epsilon=e)
    set_force_gravity_pair(c2, c3, G=G, epsilon=e)

    set_force_viscous(c1, gamma=1000)

    #c1.force += lambda t: - gamma * c1.vel
    #c2.force += lambda t: - gamma * c2.vel
    #c3.force += lambda t: - gamma * c3.vel

    world.add([c1, c2, c3])
    world.run()



from __future__ import division
from math import trunc
import ctypes
import sdl
import sdl.sdlgfx as gfx
from FGAme.core.screen import Canvas
from FGAme.input import Input
from FGAme.mainloop import MainLoop
from FGAme.draw import Color

#
# Module constants
#
HAS_INIT_SDL_VIDEO = False


class SDL2Canvas(Canvas):

    '''Implementa a interface Screen, utilizando a biblioteca Pygame'''

    # Local class-based caches
    _gfx = gfx
    _sdl = sdl

    def show(self):
        # Init video
        global HAS_INIT_SDL_VIDEO
        if not HAS_INIT_SDL_VIDEO:
            self._no_error(sdl.SDL_Init(sdl.SDL_INIT_VIDEO))
        else:
            HAS_INIT_SDL_VIDEO = True

        # Create window
        self._window = self._no_null(
            sdl.SDL_CreateWindow(
                b"FGAme App",
                sdl.SDL_WINDOWPOS_CENTERED,
                sdl.SDL_WINDOWPOS_CENTERED,
                self.width,
                self.height,
                sdl.SDL_WINDOW_SHOWN))

        # Create a renderer
        flags = sdl.SDL_RENDERER_SOFTWARE
        self._renderer = self._no_null(
            sdl.SDL_CreateRenderer(self._window, -1, flags))

        # Create a rect instance that stores the window shape
        self._screen_rect = self._no_null(
            sdl.SDL_Rect(x=0, y=0, w=self.width, h=self.height))

        # Saves background color
        if self.background is not None:
            R, B, G, A = Color(self.background)
            self._bg_color = (R << 24) + (G << 16) + (B << 8) + A
        else:
            self._bg_color = None
        self._LP_short = gfx.aapolygonRGBA.argtypes[1]

    def flip(self):
        sdl.SDL_RenderPresent(self._renderer)

    def _map_point(self, point):
        x, y = point
        return (trunc(x), trunc(self.height - y))

    def _sdl_error(self):
        msg = sdl.SDL_GetError()
        raise RuntimeError('SDL error: %s' % msg)

    def _no_error(self, value):
        if value != 0:
            msg = sdl.SDL_GetError()
            raise RuntimeError('SDL error (%s): %s' % (value, msg))

    def _no_null(self, value):
        if not value:
            msg = sdl.SDL_GetError()
            raise RuntimeError('SDL error (%s): %s' % (value, msg))
        return value

    #
    # Desenho de figuras geométricas primitivas
    #
    def draw_raw_aabb_solid(self, aabb, color):
        Y = self.height
        R, G, B, A = color
        if gfx.boxRGBA(
                self._renderer,
                trunc(aabb.xmin),
                trunc(Y - aabb.ymax),
                trunc(aabb.xmax),
                trunc(Y - aabb.ymin),
                R, G, B, A) != 0:
            self._sdl_error()

    def draw_raw_circle_solid(self, circle, color):
        self._no_error(
            gfx.filledCircleRGBA(
                self._renderer,
                trunc(circle.x),
                trunc(self.height - circle.y),
                trunc(circle.radius), *color))

        self._no_error(
            gfx.aacircleRGBA(
                self._renderer,
                trunc(circle.x),
                trunc(self.height - circle.y),
                trunc(circle.radius), *color))

    def draw_raw_circle_border(self, circle, width, color):
        safe_operator = self._no_error
        x, y = trunc(circle.x), trunc(self.height - circle.y)
        renderer = self._renderer

        if width == 1:
            radius = trunc(circle.radius)
            safe_operator(gfx.aacircleRGBA(renderer, x, y, radius, *color))
        else:
            # Very ugly hack! can it be better?
            min_r = trunc(circle.radius - width / 2)
            max_r = trunc(circle.radius + width / 2)
            for r in range(min_r, max_r):
                safe_operator(gfx.aacircleRGBA(renderer, x, y, r, *color))

    def _get_poly_xy(self, poly):
        '''Implementação comum que cria lista de posições x e y para passsar
        para draw_raw_poly_solid e draw_raw_poly_border'''
        N = len(poly)
        height = self.height
        int16_array = ctypes.c_int16 * N
        Xc = int16_array()
        Yc = int16_array()
        for i, pt in enumerate(poly):
            Xc[i] = trunc(pt.x)
            Yc[i] = trunc(height - pt.y)
        return Xc, Yc

    def draw_raw_poly_solid(self, poly, color):
        N = len(poly)
        X, Y = self._get_poly_xy(poly)
        color = Color(color)
        self._no_error(
            gfx.aapolygonRGBA(
                self._renderer, X, Y, N, *color))
        self._no_error(
            gfx.filledPolygonRGBA(
                self._renderer, X, Y, N, *color))

    def draw_raw_segment(self, segment, width, color):
        height = self.height
        pt1, pt2 = segment
        if width == 1:
            self._no_error(
                gfx.aalineRGBA(
                    self._renderer,
                    trunc(pt1.x),
                    trunc(height - pt1.y),
                    trunc(pt2.x),
                    trunc(height - pt2.y),
                    *color))
        else:
            self._no_error(
                gfx.thickLineRGBA(
                    self._renderer,
                    trunc(pt1.x),
                    trunc(height - pt1.y),
                    trunc(pt2.x),
                    trunc(height - pt2.y),
                    trunc(width),
                    *color))

    def draw_raw_poly_border(self, poly, width, color):
        # TODO: desenhar linhas espessas
        N = len(poly)
        X, Y = self._get_poly_xy(poly)
        self._no_error(
            gfx.aapolygonRGBA(
                self._renderer, X, Y, N, *Color(color)))

    def clear_background(self, color=None):
        renderer = self._renderer
        sdl.SDL_RenderClear(renderer)

        ret = 0
        if color is None and self._bg_color is not None:
            ret = gfx.boxColor(
                renderer, 0, self.height, self.width, 0, self._bg_color)
        else:
            R, G, B, A = Color(color)
            ret = gfx.boxRGBA(
                renderer, 0, self.height, self.width, 0, R, G, B, A)

        if ret != 0:
            msg = sdl.SDL_GetError()
            raise RuntimeError('SDL error: %s' % msg)


class SDL2Input(Input):

    '''Objetos do tipo listener.'''

    #=========================================================================
    # Conversões entre strings e teclas
    #=========================================================================
# Setas direcionais
#     KEY_CONVERSIONS = {
#         'up': SDLK_UP, 'down': SDLK_DOWN, 'left': SDLK_LEFT, 'right': SDLK_RIGHT,
#         'return': SDLK_RETURN, 'space': SDLK_SPACE, 'enter': SDLK_RETURN,
#     }
#
# Adiciona as letras e números
#     chars = '0123456789' + string.ascii_lowercase
#     for c in chars:
#         KEY_CONVERSIONS[c] = getattr(sdl, 'SDLK_' + c)

    #
    # Laço principal de escuta de eventos
    #

    def poll(self):
        for event in sdl_ext.get_events():
            if event.type == sdl.SDL_QUIT:
                raise SystemExit
            elif event.type == sdl.SDL_KEYDOWN:
                self.on_key_down(event.key.keysym.sym)
            elif event.type == sdl.SDL_KEYUP:
                self.on_key_up(event.key.keysym.sym)
            elif event.type == sdl.SDL_MOUSEMOTION:
                # TODO: converter para coordenadas locais em screen
                # self.on_mouse_motion(event.pos)
                pass

        # self.on_long_press()


class SDL2MainLoop(MainLoop):

    '''The SDL2 main loop'''



import FGAme
from FGAme.mathtools import asvector


class TileManager(object):

    '''Configura um tileset que permite dispor objetos no mundo'''

    def __init__(self, shape=(50, 50), origin=(0, 0)):
        self.shape = asvector(shape)
        self.origin = asvector(origin)
        self.tiles = []
        self.specs = {}
        self.charspecs = {}

    def register_spec(self, name, char, tile=None, **kwds):
        '''
        Registra uma especificação de ladrilho associando a um determinado
        nome e caractere. Por padrão, o ladrilho é uma AABB preta do formado
        especificado pelo atributo shape.

        A função aceita vários argumentos adicionais que determinam as
        propriedades do ladrilho.


        Parameters
        ----------

        tile :
            Se especificado, corresponde ao ladrilho que será replicado a cada
            comando :meth:`add_tile` ou :meth:`add_tileset`.
        mode : 'aabb', 'circle'


        Todos os argumentos opcionais da classe :cls:`FGAme.objects.Object`
        são aceitos.
        '''

        if tile is None:
            tile = self._make_tile(name, **kwds)
        self.specs[name] = tile
        if char is not None:
            self.charspecs[char] = tile

    def _make_tile(self, name, mode='aabb', **kwds):
        '''Cria ladrinho para o método register_spec'''

        modes = {
            'aabb': {
                'cls': FGAme.AABB,
                'shape': self.shape,
            },
            'circle': {
                'cls': FGAme.Circle,
                'radius': min(self.shape) / 2,
            },
            'poly': {
                'cls': FGAme.Poly,
            },

        }

        # Obtem argumentos e classe de inicialização a partir do modo
        mode = modes[mode.lower()]
        cls = mode.pop('cls')
        kwds.update(mode)

        # Configura argumentos padrão para todos os objetos
        kwds.setdefault('mass', 'inf')

        # Inicializa e reposiciona
        tile = cls(**kwds)
        tile.name = name
        tile.move(self.origin - tile.pos_se)

        return tile

    def add_tile(self, pos, tile):
        '''Adiciona o ladrilho especificado na posição dada.

        O ladrilho pode ser um objeto arbitrário que pode ser adicionado a um
        mundo ou pode ser especificado pelo nome ou caractere de identificação.

        A posição (i, j) corresponde ao valor (normalmente) inteiro da posição
        medida em número de ladrilhos deslocados nas coordenadas x e y
        respectivamente. Deste modo, um valor fracionário como (10, 0.5)
        significa deslocar 10 ladrilhos para a direita e meio ladrilho para
        cima a partir da origem definida em `self.origin`.
        '''

        i, j = pos
        dx, dy = self.shape
        x0, y0 = self.origin

        if isinstance(tile, str):
            name = tile
            try:
                if len(name) == 1:
                    tile = self.charspecs[name]
                else:
                    tile = self.specs[name]
            except KeyError:
                raise ValueError('%r is not a valid tile name' % name)

        tile = tile.copy()
        tile.move(x0 + i * dx, y0 + j * dy)
        self.tiles.append(tile)

    def add_tileset(self, data):
        '''Adiciona um tileset completo a partir da string de especificação.

        Um exemplo de string ``data`` é dado abaixo::
            |             **
            |        **     xxxxxxxxxx
            |           xxx
            | ##    xxx
            | H#                         ^^
            |xxxxxxxxxxxxxxxxxxxx    xxxxxxxxxxx
            |xxxxxxxxxxxxxxxxxxxx^^^^xxxxxxxxxxx

        Cada linha deve começar com o caracter "|". A partir daí a posição de
        cada ladrilho é calculada para corresponder à mesma posíção na string.
        Os ladrilhos são então inseridos por caractere de identificação (o
        espaço em branco significa a ausência de ladrilho).

        Neste caso, temos o chão quase inteiramente formado por ladrilhos do
        tipo "x" e um ladrilho do tipo "H" na posição (1, 2). Os outros
        ladrilhos estão dispostos de maneira análoga.

        O usuário pode utilizar qualquer caractere unicode à exceção de espaço,
        tab, newline e hashtag para identificar um ladrilho. Espaços em branco
        são interpretados como a ausência de ladrilhos e as linhas representam
        cada camada do ladrilhamento. O hashtag ("#") é ignorado da mesma forma
        que o espaço. Normalmente é utilizado para sinalizar que a posição está
        tomada por algum ladrilho vizinho. No caso, o ladrilho H pode ocupar um
        tamanho de 2x2, apesar de estar posicionado somente em uma casa
        específica.

        Cada ladrilho deve ter sido criado previamente pela função
        :meth:`register_spec`. Caso o usuário utilize um caractere inválido,
        isto corresponderá a um erro.
        '''

        y = data.count('|') - 1
        for line in data.splitlines():
            line = line.lstrip()
            if line.startswith('|'):
                line = line[1:]
            else:
                continue

            for x, char in enumerate(line):
                if char not in ' #':
                    self.add_tile((x, y), char)
            y -= 1

    def update_world(self, world, layer=0):
        '''Usado por World.add() para adicionar o tileset'''

        for tile in self.tiles:
            if tile.is_rogue():
                world.add(tile.copy(), layer)
            else:
                raise RuntimeError('tile already present in world')

    def __iter__(self):
        return iter(self.tiles)

    def __len__(self):
        return len(self.tiles)

if __name__ == '__main__':
    from FGAme import World, AABB

    ts = '''
    |            oo
    |        oo     xxxxxxxxxxxxxxxxxx
    |           xxx
    |       xxx
    |                                        ii
    |xxxxxxxxxxxxxxxxxxxxxxxxxxxxx    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxiiiixxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    '''

    w = World()
    tm = TileManager((30, 30))
    tm.register_spec('brick', 'x', color='red')
    tm.register_spec('coin', 'o', mode='circle', color='yellow')
    tm.register_spec('spike', 'i', color='black')
    tm.add_tileset(ts)

    w.add(tm)
    w.run()



# -*- coding: utf8 -*-
from __future__ import division
import contextlib
from FGAme.mathtools import pi, Vec2, Direction2, ux2D, uy2D
from FGAme.draw import Color
from FGAme.draw import base as draw

rad = pi / 180  # Conversão de graus para radianos
black = Color('black')


class RenderTree(object):

    '''Representa uma árvore de objetos que serão desenhados na tela

    Exemplos
    --------

    Inicializamos a árvore com alguns objetos. A string 'foo' foi colocado em
    uma camada superior e portanto será renderizada por último

    >>> tree = RenderTree()
    >>> tree.add('foo', 1)
    >>> tree.add('bar')
    >>> tree.add('blaz')

    Podemos navegar na árvore utilizando o iterador `walk()`

    >>> list(tree.walk())
    ['bar', 'blaz', 'foo']

    O método remove() permite remover objetos

    >>> tree.remove('bar');
    >>> 'bar' in tree, 'foo' in tree
    (False, True)
    '''

    is_tree = True
    visible = True

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def __init__(self, parent=None):
        self._data = []
        self.parent = None

    # Métodos mágicos #########################################################
    def __contains__(self, obj):
        return any(obj in L for _, L in self._data)

    def __iter__(self):
        return self.walk()

    def __len__(self):
        return sum(len(layer) for layer in self._data)

    def __getitem__(self, idx):
        for i, obj in enumerate(self):
            if i == idx:
                return obj
        raise IndexError(idx)

    # Controle de objetos #####################################################
    def add(self, obj, layer=0):
        '''Adiciona um objeto ou um galho (outro elemento de RenderTree) na
        camada especificada'''

        for i, (layer_idx, data) in enumerate(list(self._data)):
            if layer == layer_idx:
                data.append(obj)
                break
            elif layer < layer_idx:
                self._data.insert(i, (layer, [obj]))
                break
        else:
            self._data.append((layer, [obj]))

    def remove(self, value):
        '''Remove a primeira ocorrência de um valor.'''

        for _, L in self._data:
            if value in L:
                L.remove(value)
                break
        else:
            raise ValueError('object %r not in RenderTree' % value)

    def remove_all(self, value):
        '''Remove todas as ocorrências do valor dado.'''

        for _, L in self._data:
            while value in L:
                L.remove(value)

    def walk(self, reverse=False):
        '''Percorre sobre todos os objetos na ordem correta. Se reverse=True,
        percorre os objetos na ordem contrária.
        '''

        if reverse:
            for _, L in reversed(self._data):
                for obj in reversed(L):
                    yield obj
        else:
            for _, L in self._data:
                for obj in L:
                    yield obj

    def iter_layers(self, skip_empty=True):
        '''Itera sobre as camadas retornando a lista de objetos em cada
        camada'''

        if skip_empty:
            for (_, L) in self._data:
                yield L

        else:
            idx = self._data[0][0]
            for i, L in self._data:
                if i == idx:
                    idx += 1
                    yield L
                else:
                    while idx < i:
                        idx += 1
                        yield []

    def get_layer(self, idx):
        '''Retorna uma lista com os objetos da i-ésima camada'''

        for i, L in self._data:
            if i == idx:
                return L
        else:
            return []

    def update(self, dt):
        '''Percorre todos os objetos na árvore invocando o método update()'''

        for obj in self.walk():
            obj.update(dt)

    def draw(self, screen):
        '''Percorre todos os objetos na árvore invocando o método
        `obj.paint(screen)`'''

        for obj in self.walk():
            if obj.visible:
                try:
                    obj.draw(screen)
                except Exception:
                    print('debug: error drawing %s' % obj)
                    raise

    def linearize(self, layer=0):
        '''Retorna uma versão linearizada da árvore de renderização onde
        todos os objetos são recolocados na mesma camada'''

        data = (layer, list(self.walk()))
        new = RenderTree(parent=self.parent)
        new._data.append(data)
        return new


class Pen(RenderTree):

    '''Uma RenderTree que constroi objetos usando uma metáfora parecida com o
    módulo turtle do Python.

    Diferentemente do resto da FGAme, utiliza graus (e não radianos) como
    medida padrão para ângulos. Isto previne erros de arredondamento em
    rotações.

    Em 64 bits, 6 rotações de pi/6 (30 graus) não resultam em uma rotação
    completa de pi devido ao acúmulo de erros de arredondamento. Frações
    menores geralmente resultam em erros de arredondamento ainda maiores.
    Quando representamos esta rotação em graus, os erros de arredondamento só
    aparecem para rotações que representam uma fração de grau, sendo bem menos
    perceptíveis para o usuário.

    Em arquiteturas de 32 bits, o resultado é pior.


    Examples
    --------

    Uma caneta desenha traços na tela

    >>> pen = Pen()
    >>> for _ in range(4):
    ...     pen.fwd(100)
    ...     pen.left(90)
    >>> pen.commit()
    >>> pen[0]
    Path([(0.0, 0.0), (100.0, 0.0), (100.0, 100.0), (0.0, 100.0), (0.0, 0.0)])

    Podemos desenhar figuras geométricas usando gerenciadores de contexto

    >>> with pen.rect():
    ...    pen.left(45)
    ...    pen.fwd(50)

    #>>> pen[1]
    #AABB([0.0, 35.4, 0.0, 35.4])



    '''

    # Vetores em direções especiais
    _SPECIAL_DIRECTIONS = {
        0: ux2D, 90: uy2D, 180: -ux2D, 270: -uy2D,
        45: Direction2(1, 1),
        135: Direction2(-1, 1),
        225: Direction2(-1, -1),
        315: Direction2(1, -1),
    }

    def __init__(self, pos=(0, 0)):
        super(Pen, self).__init__()
        self._pos_prev = self._pos = Vec2(*pos)
        self._current = [self._pos]
        self._angle = 0
        self._angle_prev = 0
        self._dirty = False
        self._fill_color = black
        self._line_color = black
        self._line_width = 1.0
        self._pen_down = True
        self._can_move_pen_z = True
        self._can_move_pen_xy = True

    #
    # Propriedades da classe
    #
    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value):
        if self._pen_down:
            self._pos, self._pos_prev = Vec2(*value), self._pos
            self._current.append(self._pos)
        else:
            self._pos = Vec2(*value)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        if self._dirty:
            self._angle, self._angle_prev = float(value), self._angle
            self._dirty = False
        else:
            self._angle = float(value)

    @property
    def theta(self):
        return self.angle * rad

    @theta.setter
    def theta(self, value):
        self.angle = value / rad

    #
    # Controle de movimento da caneta
    #
    def up(self):
        '''Levanta a caneta e para de desenhar'''

        if self._pen_down:
            self._assure_can_move_pen_z()
            self.commit()
            self._pen_down = False

    def down(self):
        '''Abaixa a caneta e desenha'''

        if not self._pen_down:
            self._assure_can_move_pen_z()
            self._current = [self.pos]
            self._pen_down = True

    def move(self, x_or_delta, y=None):
        '''Move o objeto pelo vetor de deslocamento dado'''

        if y is not None:
            return self.move(Vec2(x_or_delta, y))

        self._assure_can_move_pen_xy()
        self.pos += x_or_delta

    def goto(self, x_or_delta, y=None):
        '''Vai até a posição especificada na tela'''

        if y is not None:
            return self.goto(Vec2(x_or_delta, y))

        self.move(x_or_delta - self._pos)

    def left(self, angle=90):
        '''Rota à esquerda pelo ângulo especificado em graus'''

        self.angle += angle

    def right(self, angle=90):
        '''Rota à direita pelo ângulo especificado em graus'''

        self.left(-angle)

    def forward(self, length):
        '''Move para frente pelo comprimento especificado em pixels'''

        self.move(self.direction() * length)

    def fwd(self, length):
        '''Atalho para `obj.forward(length)`'''

        self.forward(length)

    def backwards(self, length):
        '''Move para trás pelo comprimento especificado em pixels'''

        self.forward(-length)

    def back(self, length):
        '''Atalho para `obj.backwards(length)`'''

        self.backwards(length)

    def direction(self):
        '''Retorna um vetor unitário na direção especificada'''

        # Casos especiais, para obter direções precisas
        try:
            return self._SPECIAL_DIRECTIONS[self._angle % 360]
        except KeyError:
            return ux2D.rotate(self.theta)

    #
    # Controle de transações
    #
    def commit(self):
        '''Insere linha que estava sendo desenha na árvore'''

        if len(self._current) > 1:
            self.add(
                draw.Path(self._current, width=self._line_width,
                          color=self._fill_color))
        self.clear()

    def clear(self):
        '''Remove última linha que estava sendo desenhada'''

        self._current = [self._pos]

    #
    # Context managers
    #
    @contextlib.contextmanager
    def lock_pen_z(self):
        '''Tranca o movimento da caneta no eixo z'''

        # Set up
        can_move = self._can_move_pen_z
        self._can_move_pen_z = False

        # Enter
        yield

        # Clean
        self._can_move_pen_z = can_move

    @contextlib.contextmanager
    def lock_pen_xy(self):
        '''Tranca o movimento da caneta no plano'''

        # Set up
        can_move = self._can_move_pen_xy
        self._can_move_pen_xy = False

        # Enter
        yield

        # Clean
        self._can_move_pen_xy = can_move

    @contextlib.contextmanager
    def rect(self, line_color=None, **kwds):
        '''Gerenciador de contexto que inicia o desenho de uma AABB cujos
        limites se encontram no ponto inicial e no ponto final do caminho.'''

        # TODO: fatorar para mover implementação comum a outras figuras
        # geométricas para o mesmo método.

        # Set up
        start = self.pos
        self.commit()
        self.down()
        if line_color is None:
            line_color = self._line_color

        # Enter
        with self.lock_pen_z():
            yield

        # Clean
        end = self.pos
        self.clear()
        limits = sorted([v.x for v in [start, end]])
        limits.extend(sorted([v.y for v in [start, end]]))
        self.add(draw.AABB(*limits, line_color=line_color, **kwds))

    def circle(self, **kwds):
        '''Gerenciador de contexto que inicia o desenho de um círculo cujo
        centro está no ponto inicial e o raio vai até o ponto final'''

        raise NotImplementedError

    def aabb(self, **kwds):
        '''Gerenciador de contexto que inicia o desenho de uma AABB que
        envolve todos os pontos da trajetória realizada.'''

    def cbb(self, **kwds):
        '''Gerenciador de contexto que desenha o menor círculo que envolve
        todos os pontos da trajetória realizada'''

        raise NotImplementedError

    def poly(self, **kwds):
        '''Gerenciador de contexto que inicia o desenho de um polígono
        fechado'''

        raise NotImplementedError

    #
    # Utility private methods
    #
    def _assure_can_move_pen_z(self):
        if not self._can_move_pen_z:
            raise RuntimeError('pen is locked to vertical movement')

    def _assure_can_move_pen_xy(self):
        if not self._can_move_pen_xy:
            raise RuntimeError('pen is locked to horizontal movement')

    def _solid_kwds(self, kwds=None):
        kwds = kwds or {}
        solid = kwds.setdefault('solid', self._fill_color is not None)
        if solid:
            kwds.setdefault('color', self._fill_color)
        kwds.setdefault('line_width', self._line_width)
        kwds.setdefault('line_color', self._line_color)
        return kwds

    def _curve_kwds(self, kwds=None):
        kwds = kwds or {}
        kwds.setdefault('width', self._line_width)
        kwds.setdefault('color', self._line_color)
        return kwds

    def _add_first(self, obj):
        '''Adiciona obj e retorna seu valor'''

        self.add(obj)
        return obj

    #
    # Adiciona objetos
    #
    def _add_solid(self, constructor, args, kwds):
        '''Worker para as funções do tipo add_*solid*()'''

        kwds.setdefault('pos', self.pos)
        kwds = self._solid_kwds(kwds)
        return self._add_first(constructor(*args, **kwds))

    def add_aabb(self, *args, **kwds):
        '''Adiciona um círculo ao desenho'''

        return self._add_solid(draw.AABB, args, kwds)

    def add_circle(self, radius, pos=None, **kwds):
        '''Adiciona um círculo ao desenho'''

        kwds = self._solid_kwds(kwds)
        return self._add_first(draw.Circle(radius, pos or self.pos, **kwds))

    def add_poly(self, vertices, pos=None, **kwds):
        '''Adiciona um polígono ao desenho'''

        return self._add_first(draw.Poly(vertices, pos or self.pos, **kwds))

    def add_rectangle(self, *args, **kwds):
        '''Adiciona um retângulo ao desenho'''

        return self._add_solid(draw.Rectangle, args, kwds)

    def add_triangle(self, *args, **kwds):
        '''Adiciona um triângulo ao desenho'''

        return self._add_solid(draw.Triangle, args, kwds)

    def add_segment(self, p1, p2=None, **kwds):
        '''Adiciona um segmento de reta ao desenho.

        Se somente um ponto for fornecido, utiliza a posição inicial como
        ponto inicial e o ponto dado como posição final.'''

        if p2 is None:
            p1, p2 = self.pos, p1
        self._add(draw.Segment(p1, p2, **kwds))

    def add_ray(self, arg1, direction=None, **kwds):
        '''Adiciona um raio (reta semi-finita) ao desenho.

        Pode ser chamada como `pen.add_ray(direction)` para iniciar um raio a
        partir do ponto inicial ou `pen.add_ray(point, direction) para
        especificar tanto o ponto inicial como a direção. '''

        if direction is None:
            p0, direction = self.pos, arg1
        else:
            p0 = arg1
        self._add(draw.Ray(p0, direction, **kwds))

    def add_line(self, arg1, direction=None, **kwds):
        '''Adiciona uma reta infinita ao desenho. A assinatura é igual à
        função `pen.add_ray()`'''

        if direction is None:
            p0, direction = self.pos, arg1
        else:
            p0 = arg1
        return self._add(draw.Line(p0, direction, **kwds))

    def add_path(self, *points, **kwds):
        '''Desenha uma linha seguindo a sequencia de pontos partindo do ponto
        atual.

        Pode ser chamada com um único argumento posicional que representa uma
        sequência de pontos ou por vários argumentos que representam pontos
        individuais. Se o primeiro argumento for None ou uma Elipisis (...),
        adiciona o ponto atual ao começo da lista.
        '''

        # Checa entradas
        if not points:
            raise ValueError('must pass at least 2 points')
        elif len(points) == 1:
            return self.add_path(*points[0], **kwds)

        if points[0] is None or points[0] is Ellipsis:
            points = list(points)
            points[0] = self.pos
        return self._add(draw.Path(points, **kwds))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import FGAme
from FGAme.tests import unittest
FGAme.conf.DEBUG = True


class DemoBase:
    path = None
    maxiter = 120
    check_final_condition = None
    get_initial_condition = None
    kwds = {}

    def get_world_object(self):
        '''Load world object from path.

        Prevents execution of the run() method when loading a module.'''

        old_run = FGAme.World.run
        FGAme.World.run = lambda self: None

        try:
            mod_path, _, clsname = self.path.rpartition('.')
            mod = __import__(mod_path, fromlist=(clsname,))
            cls = getattr(mod, clsname)
        finally:
            FGAme.World.run = old_run

        if isinstance(cls, type):
            return cls(**self.kwds)
        else:
            return cls

    def run_world(self, world):
        world.run(maxiter=self.maxiter, wait=False)

    def test_world_starts(self):
        self.get_world_object()

    def test_world_runs(self):
        world = self.get_world_object()
        self.run_world(world)

    def test_world_condition(self):
        if self.check_final_condition:
            world = self.get_world_object()
            ICs = self.get_initial_condition(world)
            self.run_world(world)
            self.check_final_condition(world, ICs)


class TestGravityTutorial(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.tutorial_gravity.GravityWorld'


class TestSimpleGas(DemoBase, unittest.TestCase):
    maxiter = 60
    path = 'FGAme.demos.simulations.gas_simple.Gas'


class TestGasAABBs(DemoBase, unittest.TestCase):
    maxiter = 60
    path = 'FGAme.demos.simulations.gas_aabbs.world'


class TestGasPolys(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.gas_polys.world'


class TestSimpleCircle(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.simple_circle.world'


class TestSimpleAABB(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.simple_aabb.world'


class TestSimplePoly(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.simple_poly.world'


class TestSimpleMixed(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.simple_mixed.world'


class TestTripleCircleCollision(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.triple_circle.world'


class TestTriplePolyCollision(DemoBase, unittest.TestCase):
    path = 'FGAme.demos.simulations.triple_poly.world'

if __name__ == '__main__':
    unittest.main('FGAme.tests.demos')



# -*- coding: utf8 -*-
from FGAme import *

# Cria mundo e objetos
world = World()
world.add_bounds(width=10, use_poly=True)
p1 = RegularPoly(5, length=50, pos=(50, 300), vel=(800, 0))
p2 = RegularPoly(5, length=30, pos=(420, 320))
p3 = RegularPoly(5, length=30, pos=(420, 280))

# Insere objetos e inicia rastreio de energia
world.register_energy_tracker()
world.add([p1, p2, p3])
world.run()



from FGAme.core.screen import Canvas

class PyGletCanvas(Canvas):
    '''Implementa a interface Screen, utilizando a biblioteca Pygame'''

    def __init__(self, width, height, **kwds):
        global WINDOW
        
        super(PyGletScreen, self).__init__(width, height, **kwds)
        self._window = pyglet.window.Window(width, height)
        WINDOW = self._window

    def show(self):
        self._window.flip()

    def _map_point(self, point):
        x, y = super(PyGletScreen, self)._map_point(point)
        X, Y = self.width, self.height
        x += 0.5 * X
        y = 0.5 * Y - y
        return (trunc(x), trunc(y))

    #===========================================================================
    # Desenho
    #===========================================================================
    def draw_circle(self, pos, radius, color=(0, 0, 0), solid=True):
        pos = self._map_point(pos)
        #pygame.draw.circle(self._screen, color, pos, trunc(radius))

    def draw_poly(self, points, color=(0, 0, 0), solid=True):
        points = [ self._map_point(pt) for pt in points ]
        #pygame.draw.polygon(self._screen, color, points)

    def draw_rect(self, pos, shape, color=(0, 0, 0), solid=True):
        x, y = self._map_point(pos)
        dx, dy = shape
        dx = trunc(dx * self.zoom)
        dy = trunc(dy * self.zoom)
        y -= dy
        #pygame.draw.rect(self._screen, color, (x, y, dx, dy))
        pyglet.graphics.draw(2, pyglet.gl.GL_POINTS,
                             ('v2i', (10, 15, 30, 35))
        )

    def draw_line(self, pt1, pt2, color=(0, 0, 0), solid=True):
        raise NotImplementedError

    def clear(self, color=None):
        #color = color or self.background
        #self._screen.fill(color)
        self._window.clear()
        R, G, B = self.background
        pyglet.gl.glClearColor(R / 255., G / 255., B / 255., 1)

        win = self._window
        verts = [ (win.width * 0.9, win.height * 0.9),
                  (win.width * 0.5, win.height * 0.1),
                  (win.width * 0.1, win.height * 0.9), ]
        colors = [ (255, 000, 000),
                   (000, 255, 000),
                   (000, 000, 255), ]
        glBegin(GL_TRIANGLES)
        for idx in range(len(verts)):
            glColor3ub(*colors[idx])
            glVertex2f(*verts[idx])
        glEnd()        
        '''
Created on 19/03/2015

@author: chips
'''



#-*- coding: utf8 -*-

from FGAme.extra.declarations_parsing import parse_data, Group, Declaration, Assignment
import FGAme

# TODO: fazer um processo de filtragem das classes permitidas
DECLARATIONS_FGAME = ['AABB', 'Triangle', 'Poly']
DECLARATIONS_MAPPING = {}
DECLARATIONS_MAPPING.update({ k: 'FGAme.' + k for k in DECLARATIONS_FGAME })


def load_tree(tree):
    out = []
    for obj in tree:
        if isinstance(obj, Declaration):
            kwds = { c.name: c.value for c in obj.children }
            obj_type = getattr(FGAme, obj.name)
            out.append(obj_type(**kwds))
            
        elif isinstance(obj, Group):
            raise NotImplementedError

    return out

if __name__ == '__main__':
    example = (
        'Circle:\n'
        '    radius: 30\n'
        '    pos: (400, 300)\n'
    )
    tree = parse_data(example)
    print(load_tree(tree))


import os

if __name__ == '__main__':
    for file in ['aabb_friction', 'adamp', 'gas_aabbs', 'embolo',
                 'poly_simple', 'poly_simple2', 'gas_polys']:

        code = os.system('python %s.py' % file)
        if code != 0:
            raise SystemExit(code)



# -*- coding: utf8 -*-
'''
Módulos da FGAme
================

Configuração (FGAme.conf)
-------------------------

.. automodule:: FGAme.conf
    :members:
    :member-order: bysource

Objetos e mundo
---------------


Física e colisões
-----------------


Funções matemáticas
-------------------


Tópicos avançados
=================

Eventos
-------

.. automodule:: FGAme.events
    :members: EventDispatcher, signal, listen


Anatomia de uma colisão
-----------------------

Loop principal
--------------

'''

from FGAme.meta import __version__, __author__
from FGAme import conf
from FGAme import mathtools as math
from FGAme.input import *
from FGAme.mathtools import Vec, pi
from FGAme import draw
from FGAme.events import listen, signal, EventDispatcher
from FGAme.core import *
from FGAme import physics
from FGAme.physics import *
from FGAme.world import World
from FGAme.objects import *
from FGAme.app import *
from FGAme.extra.orientation_objects import *
from FGAme.mainloop import *
del MainLoop

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from __future__ import division
import functools
import six
from FGAme.draw import AABB
from FGAme.mathtools import asvector

try:
    import PIL.Image
except ImportError:
    from warnings import warn
    warn('PIL not found. FGAme will not be able to render images and textures')


if six.PY3:
    def lru_cache(func):
        return functools.lru_cache(maxsize=512)(func)
else:
    def lru_cache(func):
        D = {}

        @functools.wraps(func)
        def decorated(*args):
            try:
                return D[args]
            except KeyError:
                result = func(*args)
                while len(D) > 512:
                    D.popitem()
                return result
        return decorated


@lru_cache
def get_texture(path, theta=0, scale=1):
    if theta == 0 and scale == 1:
        return Texture(path)
    else:
        return get_texture(path).rotate(theta).rescale(scale)


class Texture(object):

    def __init__(self, path):
        self._pil = PIL.Image.open(path)
        self._pil.load()
        self.path = path

    def get_pil_data(self):
        '''Retorna um objeto do tipo PIL.Image.Image'''

        return self._pil

    def set_data(self, value):
        self.data = value

    def get_data(self, value):
        try:
            return self.data
        except AttributeError:
            raise RuntimeError('data attribute is not defined')

    @property
    def shape(self):
        return (self._pil.width, self._pil.height)

    @property
    def mode(self):
        return self._pil.mode


class Image(AABB):

    '''Define uma imagem/pixmap não-animado'''

    # manager = PyGameTextureManager()
    canvas_primitive = 'image'

    def __init__(self, path_or_texture, pos=(0, 0), reference='center'):
        if isinstance(path_or_texture, Texture):
            self.texture = path_or_texture
        else:
            self.texture = get_texture(path_or_texture)

        shape = asvector(self.texture.shape)
        if reference == 'center':
            pos = asvector(pos)
        elif reference == 'origin':
            pos = pos + self.shape / 2
        else:
            raise ValueError('invalid reference: %r' % reference)

        super(AABB, self).__init__(pos=pos, shape=shape)

    def draw(self, screen):
        screen.draw_image(self)



# -*- coding: utf8 -*-
from FGAme import *

# Cria mundo e objetos
world = World()
world.add_bounds(width=10)
obj1 = RegularPoly(N=3, length=130, pos=(200, 300),
                   vel=(200, 200), color='random', omega=2.2)
obj2 = AABB(shape=(100, 80), pos=(200, 450), color='random', mass='inf')
obj3 = Circle(40, pos=(600, 300), color='random', mass='inf')
obj4 = Circle(40, pos=(450, 100), color='random')

# Insere objetos e inicia rastreio de energia
world.register_energy_tracker()
world.add([obj1, obj2, obj3, obj4])
world.run()



# -*- coding: utf8 -*-
# TODO: manter este modelo de renderização?
import copy
from FGAme.draw import Color
from FGAme.mathtools import shapes

black = Color('black')
white = Color('white')


class Drawable(object):

    '''Classe mãe para todos os objetos que podem ser desenhados na tela'''

    __slots__ = []

    visible = True

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def copy(self):
        '''Retorna uma cópia do objeto'''

        return copy.copy(self)

    def draw(self, canvas):
        '''Desenha objeto na tela'''

        try:
            func = getattr(canvas, self._canvas_shape_name)
        except AttributeError:
            name = 'draw_' + type(self).__name__.lower()
            type(self)._canvas_shape_name = name
            func = getattr(canvas, self._canvas_shape_name)
        return func(self, **self._draw_kwds())


###############################################################################
def lazy_curve_init(self, *args, **kwds):
    '''Generic init implementation for multiple inheritance from a
    mathematical shape and Curve.

    It is less efficient that it could be, but is a workable default for
    lazy coders.'''

    get = kwds.pop
    Curve.__init__(self, get('width', 1.0), get('color', black))
    type(self).__bases__[0].__init__(self, *args, **kwds)


class Curve(Drawable):

    '''Classe pai para todos os objetos geométricos que definem uma curva
    aberta.
     '''

    def __init__(self, width=1.0, color=black):
        self._color = Color(color)
        self.width = width

    def _init_colors_from_kwds(self, kwds):
        '''Extrai as keywords do dicionário kwds, e inicializa as propriedades
        de cores do objeto. As outras entradas do dicionário podem ser
        passadas para o construtor do objeto geométrico'''

        get = kwds.pop
        width = get('width', 1.0)
        color = get('color', black)
        Curve.__init__(self, width, color)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if isinstance(value, Color):
            self._color = value
        else:
            self._color = Color(value)

    @property
    def line_color(self):
        return self._color

    @line_color.setter
    def line_color(self, value):
        self._color = Color(value)

    def _draw_kwds(self):
        return dict(color=self.color, width=self.width)


class Segment(shapes.mSegment, Curve):

    '''Um segmento de reta que vai do ponto start até end'''

    __init__ = lazy_curve_init


class Ray(shapes.mRay, Curve):

    '''Um segmento de reta semi-finito que inicia em um ponto start
    específico'''

    __init__ = lazy_curve_init


class Line(shapes.mLine, Curve):

    '''Uma linha infinita'''

    __init__ = lazy_curve_init


class Path(shapes.mPath, Curve):

    '''Um caminho formado por uma sequência de pontos'''

    __init__ = lazy_curve_init

# Objetos sólidos (figuras fechadas) ##########################################


def lazy_shape_init(self, *args, **kwds):
    '''Generic init implementation for multiple inheritance from a
    mathematical shape and Shape.

    It is less efficient that it could be, but is a workable default for
    lazy coders.'''

    get = kwds.pop
    Shape.__init__(self, get('solid', True), get('color', black),
                   get('line_width', 0.0), get('line_color', black))
    type(self).__bases__[0].__init__(self, *args, **kwds)


class Shape(Drawable):

    '''Classe pai para todas as figuras fechadas. Define as propriedades
    solid, color, line_width e line_color.'''

    def __init__(self, solid=True, color=black,
                 line_width=0.0, line_color=black):

        if not isinstance(color, Color):
            color = Color(color)
        if not isinstance(line_color, Color):
            line_color = Color(color)
        self._color = color
        self._line_color = line_color
        self.line_width = line_width
        self.solid = bool(solid)

    def _init_colors_from_kwds(self, kwds):
        '''Extrai as keywords do dicionário kwds, e inicializa as propriedades
        de cores do objeto. As outras entradas do dicionário podem ser
        passadas para o construtor do objeto geométrico'''

        solid = kwds.pop('solid', True)
        color = kwds.pop('solid', black)
        line_color = kwds.pop('solid', black)
        line_width = kwds.pop('solid', 0.0)
        Shape.__init__(self, solid, color, line_width, line_color)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = Color(value)

    @property
    def line_color(self):
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        self._line_color = Color(value)

    def _draw_kwds(self):
        return dict(
            solid=self.solid,
            color=self.color,
            line_width=self.line_width,
            line_color=self.line_color)


class Circle(shapes.mCircle, Shape):

    '''Um círculo de raio `radius` e centro em `center`.'''

    def __init__(self, radius, pos=(0, 0), **kwds):

        shapes.mCircle.__init__(self, radius, pos)
        Shape.__init__(self, **kwds)


class AABB(shapes.mAABB, Shape):

    '''Caixa de contorno alinhada aos eixos'''

    __init__ = lazy_shape_init


class Sprite(AABB):

    '''Define uma image animada. O controle de frames é feito a cada chamada
    do método draw(). Desta forma, caso o método seja chamado duas vezes por
    frame, a velocidade de animação será o dobro da desejada.'''


class Circuit(shapes.mCircuit, Shape):

    '''Um caminho fechado com um interior bem definido.'''

    __init__ = lazy_shape_init


class Poly(shapes.mPoly, Shape):

    '''Objetos da classe Poly representam polígonos dados por uma lista
    de vértices.'''

    __init__ = lazy_shape_init


class Rectangle(shapes.mRectangle, Shape):

    '''Especialização de Poly para representar retângulos.

    Pode ser inicializado como uma AABB, mas com um parâmetro extra `theta`
    para definir um ângulo de rotação.'''

    # TODO: não consegue herdar de Poly. Porque?
    __init__ = lazy_shape_init

    def draw(self, canvas):
        canvas.draw_poly(self, **self._draw_kwds())


class Triangle(shapes.mTriangle, Shape):

    '''Especialização de Poly para representar triângulos.'''

    # TODO: não consegue herdar de Poly. Porque?
    __init__ = lazy_shape_init

    def draw(self, canvas):
        canvas.draw_poly(self, **self._draw_kwds())


if __name__ == '__main__':
    seg = Segment((0, 1), (1, 0), width=2, color=white)
    print(seg.__dict__)



# -*- coding: utf8 -*-
from FGAme import *


# Cria mundo e objetos
world = World()
world.add_bounds(width=10)
obj1 = Circle(50, pos=(200, 330), vel=(300, 0), color='random')
obj2 = Circle(60, pos=(200, 450), color='random')
obj3 = Circle(30, pos=(600, 300), color='random', mass='inf')

# Insere objetos e inicia rastreio de energia
world.register_energy_tracker()
world.add([obj1, obj2, obj3])
world.run()



# -*- coding: utf8 -*-
import types

__all__ = ['lazy', 'delegate_to', 'autodoc', 'popattr']


class lazy(object):

    '''Implementa uma propriedade "preguiçosa": ela é calculada apenas durante o
    primeiro uso e não durante a inicialização do objeto.'''

    def __init__(self, force):
        self.force = force

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        value = self.force(obj)
        setattr(obj, self.force.__name__, value)
        return value


class delegate_to(property):

    '''Sincroniza com um attributo de contido em um sub-attributo da classe.

    Exemplo
    -------

    >>> class Foo(object):
    ...     def __init__(self):
    ...         self._data = []
    ...
    ...     add = delegate_to('_data.append')

    Agora criamos um objeto

    >>> x = Foo()
    >>> x.add(1); x.add(2); x._data
    [1, 2]
    '''

    def __init__(self, delegate, read_only=False):
        self.delegate = delegate
        attrs = delegate.split('.')
        delegate = attrs.pop(0)

        if len(attrs) == 0:
            def fget(self):
                return getattr(self, delegate)

            def fset(self, value):
                setattr(self, delegate, value)

            def fdel(self):
                delattr(self, delegate)

        elif len(attrs) == 1:
            attr = attrs[0]

            def fget(self):
                delegate_obj = getattr(self, delegate)
                return getattr(delegate_obj, attr)

            def fset(self, value):
                delegate_obj = getattr(self, delegate)
                setattr(delegate_obj, attr, value)

            def fdel(self):
                delegate_obj = getattr(self, delegate)
                delattr(delegate_obj, attr)

        else:
            raise NotImplementedError('more than one dot')

        if read_only:
            super(delegate_to, self).__init__(fget)
        else:
            super(delegate_to, self).__init__(fget, fset, fdel)


def autodoc(cls):
    '''Decorador de classe que insere automaticamente as strings de
    documentação nos métodos não-documentados de uma classe utilizando a
    string da classe mãe'''

    func_t = types.FunctionType

    for attr, value in vars(cls).items():
        if isinstance(value, func_t) and not value.__doc__:
            for sub in cls.mro():
                try:
                    doc = getattr(sub, attr).__doc__
                except AttributeError:
                    doc = None
                    break
                if doc:
                    break

            value.__doc__ = doc

    return cls


def popattr(obj, attr, value=None):
    '''Returns attribute `attr` from `obj` and delete it afterwards.
    If attribute does not exist, return `value`'''

    try:
        result = getattr(obj, attr)
    except AttributeError:
        return value
    else:
        delattr(obj, attr)
        return result


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf8 -*-
from FGAme import *
from random import uniform, choice


class Pong(World):

    def __init__(self, **kwds):
        # Inicializa o mundo
        H, W = self.screen_height / 2, self.screen_width / 2
        super(Pong, self).__init__(restitution=0.9, gravity=self.gravity)
        self.add_bounds(10, 3 * W, 10, 2 * H - 10, delta=400)
        self.listen('frame-enter', self.check_fail)
        self.listen('frame-enter', self.update_time)
        self.listen('frame-enter', self.accelerate_ball)
        self.listen('long-press', 'up', self.move_up)
        self.listen('long-press', 'down', self.move_down)
        self.listen('key-down', 'left', self.move_left)

        # Cria a raquete
        Y = self.pong_height
        X = self.pong_width
        self.pong = pong = Poly([(0, 0), (0, Y), (-X / 3, Y),
                                 (-X, 0.66 * Y), (-X, 0.33 * Y), (-X / 3, 0)])
        pong.make_static('angular')
        pong.move((2 * W - 50, H - Y / 2))
        pong.listen('collision', self.pong_collision)
        pong.mass *= 5
        pong.force = lambda t: -1000000 * Vec2(pong.pos.x - self.pong_x, 0)
        pong.damping = 10
        self.pong_x = self.pong.pos.x

        # Cria a barra de tempo
        self.timebar = AABB(shape=(20, 20), pos=(2 * W - 10, 20),
                            color=(255, 225, 0))

        # Adiciona objetos
        self.ball = self.new_ball()
        self.add([self.pong, self.ball])
        self.add(self.timebar, layer=1)
        self.obstacle = []
        for i in range(self.obstacle_N):
            new = self.new_obstacle()
            self.obstacle.append(new)
            self.add(new)

        # Define amplitude de movimento do pong
        self.max_pong_y = 2 * H - self.pong_height / 2 - 10
        self.min_pong_y = self.pong_height / 2 + 10

        # Desenha pontos de hit acinzentados
        self.hits = 0
        for i in range(self.max_hits):
            self.make_hit_mark(i, self.hits_bg)

        # Salva parâmetros adicionais
        for k, v in kwds.items():
            setattr(self, k, v)

    # Variáveis ###############################################################
    # Formas, tamanhos e posições
    screen_height = 600
    screen_width = 800
    pong_height = 150
    pong_width = 20
    ball_sides = 5
    ball_size = 30
    hit_size = 10
    pong_step = 7
    inertia_multiplier = 20

    # Velocidades
    ball_speed = 400
    max_ball_speed = 500

    # Lógica do jogo
    max_hits = 12
    obstacle_N = 15
    obstacle_sides = 4
    obstacle_size = 30
    obstacle_color = (50, 50, 100)
    obstacle_dynamic = True

    # Cores
    hits_bg = (150, 150, 150)
    gravity = 0
    next_params = {}

    # Criação e inicialização de objetos ######################################
    def new_ball(self):
        '''Inicializa o a bola'''

        H, W = self.screen_height / 2, self.screen_width / 2
        ball = RegularPoly(self.ball_sides, self.ball_size, color='red')
        ball.inertia *= self.inertia_multiplier
        ball.move((W, H))
        V = self.ball_speed
        speed = -V, choice([-1, 1]) * uniform(0.25 * V, 2 * V)
        ball.boost(speed)
        return ball

    def new_obstacle(self):
        '''Cria um objeto aleatório que fica no meio da tela'''

        # Cria obstáculo
        obj = RegularPoly(self.obstacle_sides, self.obstacle_size,
                          color=self.obstacle_color, density=1)
        obj.scale(uniform(0.75, 2))
        obj.rotate(uniform(0, 2 * math.pi))
        obj.inertia *= self.inertia_multiplier
        if not self.obstacle_dynamic:
            obj.make_static()

        # Define amplitude de movimento para posições aleatórias
        W, H = self.screen_width / 2, self.screen_height / 2
        xmin = - obj.xmin
        xmax = 2 * W - obj.xmax - 2 * self.pong_width
        ymin = - obj.ymin
        ymax = 2 * H - obj.ymax

        # Retorna objeto em nova posição
        obj.move((uniform(xmin, xmax), uniform(ymin, ymax)))
        return obj

    def make_hit_mark(self, n_hits, color='red'):
        '''Retorna um objeto que mostra na tela o número de hits'''

        W, H = self.screen_width / 2, self.screen_height / 2
        pos_x = 2 * self.hit_size - W + 3 * self.hit_size * n_hits + 10
        pos_y = H - 2 * self.hit_size - 10
        hit = draw.Circle(radius=self.hit_size, pos=(pos_x, pos_y),
                          color=color)
        self.add(hit, layer=1)

    # Callbacks de interação com o usuário ####################################
    def move_up(self):
        '''Acionado com a seta para cima'''

        if self.pong.pos.y < self.max_pong_y:
            self.pong.move((0, self.pong_step))
            self.pong.vel = (0, 400)

    def move_down(self):
        '''Acionado com a seta para baixo'''

        if self.pong.pos.y > self.min_pong_y:
            self.pong.move((0, -self.pong_step))
            self.pong.vel = (0, -400)

    def move_left(self):
        '''Executado quando o usuário aperta a seta para o lado'''

        self.pong.boost((-1000, 0))

    def update_time(self):
        '''Atualizado a cada frame para incrementar a barra de contagem do
        tempo'''

        self.timebar.ymax = ymax = self.time * 20 - self.screen_height / 2 + 20
        if ymax > (self.screen_height / 2 - 10):
            self.next()

    def accelerate_ball(self):
        '''Incrementa a velocidade da bola até um determinado limite'''

        V = self.ball.vel.norm()
        if V < self.max_ball_speed:
            V += 3
            self.ball.vel = self.ball.vel.normalize() * V

    def check_fail(self):
        '''Checa se o jogador perdeu e acrecenta um hitpoint, em caso
        positivo'''

        if self.ball.pos.x > self.screen_width + 100:
            self.hit_increment()
            self.hit_increment()
            # self.remove(self.ball)
            self.ball.move((200, 0))
            self.ball.make_static()
            self.ball = self.new_ball()
            self.add(self.ball)

        for i, obj in enumerate(self.obstacle):
            if obj.pos.x > self.screen_width + 100:
                self.hit_increment()
                # self.remove(obj)
                obj.move((200, 0))
                obj.make_static()
                del self.obstacle[i]
                break

    def hit_increment(self):
        '''Incrementa os hit counts'''
        # Testa derrota
        if self.hits >= self.max_hits - 1:
            self.loose()

        # Incrementa o número de hits
        self.make_hit_mark(self.hits)
        self.hits += 1

    def pong_collision(self, col):
        '''Chamado quando a bola colide com o pong'''

        # testa se é uma colisão com a bola ou com as paredes
        if self.ball in col:
            col.resolve()
            raise

            y_ball = self.ball.pos.y
            y_pong = self.pong.pos.y
            delta_y = y_ball - y_pong
            delta_vel = Vec2(0, delta_y * 10)

            # Edita a velocidade
            vel = self.ball.vel
            vel += delta_vel
            if abs(vel.x) > abs(0.2 * vel.y):
                vel *= self.ball.vel.norm() / vel.norm()
                self.ball.vel = vel

    def loose(self):
        '''Chamado quando o usuário perde a fase'''

        self.stop()
        game_over()

    def next(self):
        '''Chama a próxima fase'''

        self.stop()
        new = Pong(**self.next_params)
        new.run()


def game_over():
    '''Executed when the game finishes'''

    from FGAme.extra.letters import add_word

    Sx, Sy = 5, 2
    world = World(background=(255, 0, 0), gravity=10)
    letters = add_word('game over', world, scale=5, pos=(180, 350))
    for l in letters:
        l.omega = uniform(-0.1, 0.1)
        l.vel += uniform(-Sx, Sx), uniform(-Sy, Sy)
        l.inertia *= 20
    world.run()

if __name__ == '__main__':
    Pong().run()



# -*- coding: utf8 -*-
'''
Re-define os objetos do módulo FGAme.physics e adiciona propriades extras de
renderização
'''

from FGAme import conf
from FGAme import physics
from FGAme.util import lazy
from FGAme.draw import Color, color, Image
from FGAme.events import EventDispatcherMeta, signal
from FGAme.resources import resources
DEBUG = False

__all__ = ['AABB', 'Circle', 'Poly', 'RegularPoly', 'Rectangle']

black = Color('black')


@EventDispatcherMeta.decorate
class ObjectMixin(object):
    _is_mixin_ = True
    _mixin_args = ['color', 'line_color', 'line_width', 'image', 'animation',
                   'sprite']

    long_press = signal('long-press', 'key', delegate_to='_input')
    key_up = signal('key-up', 'key', delegate_to='_input')
    key_down = signal('key-down', 'key', delegate_to='_input')
    mouse_motion = signal('mouse-motion', delegate_to='_input')
    mouse_button_up = signal(
        'mouse-button-up', 'button', delegate_to='_input')
    mouse_button_down = signal(
        'mouse-button-down', 'button', delegate_to='_input')
    mouse_long_press = signal(
        'mouse-long-press', 'button', delegate_to='_input')

    @lazy
    def _input(self):
        return conf.get_input()  # @UndefinedVariable

    def __init__(self, *args, **kwds):
        mixin_kwds = self._extract_mixin_kwargs(kwds)
        self._init_physics(*args, **kwds)
        self._init_visualization(**mixin_kwds)

        # Action control
        self.visible = True
        self.actions = []
        self.to_remove = []
        self.skip_frame = False
        self.scheduled = False          # deprecated, soon to be removed
        self.scheduled_calls = []       #: list of scheduled callbacks
        self.scheduled_interval_calls = []
        self.is_running = False         #: whether of not the object is running

    def _extract_mixin_kwargs(self, kwds):
        D = {}
        mixin_args = self._mixin_args
        for k in kwds:
            if k in mixin_args:
                D[k] = kwds[k]
        for k in D:
            del kwds[k]
        return D

    def _init_visualization(
            self,
            color='black', line_color=None, line_width=1,
            image=None, animation=None, sprite=None):
        '''Init visualization parameters'''

        self._image = None
        self._sprite = None
        self._animation = None
        self._color = None
        self._line_color = None
        self._line_width = line_width
        self.color = color
        self.line_color = line_color
        self.image = image
        self.sprite = sprite
        self.animation = animation

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            self._color = None
        else:
            self._color = Color(value)

    @property
    def line_color(self):
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        if value is None:
            self._line_color = None
        else:
            self._color = Color(value)

    @property
    def line_width(self):
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = value

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        if value is None:
            self._image = None
            return

        if isinstance(value, str):
            value = resources.find_image(value)
            value = Image(value, self.pos)
        self._image = value
        self._image.offset = self._image.pos - self.pos

    @property
    def animation(self):
        return self._animation

    @animation.setter
    def animation(self, value):
        if isinstance(value, str):
            value = resources.find_animation(value)
        self._animation = value

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def draw(self, screen):
        img = self._image
        if self._image is not None:
            img.pos = self.pos + img.offset
            return img.draw(screen)
        else:
            return self.draw_shape(screen)


class AABB(ObjectMixin, physics.AABB):
    _init_physics = physics.AABB.__init__

    def draw_shape(self, screen):
        if self._color is not None:
            color = self._color
            lw, lc = self._line_width, self._line_color
            screen.draw_aabb(self.aabb, True, color, lw, lc)

        elif self._line_width:
            lw, lc = self._line_width, self._line_color
            screen.draw_aabb(self.aabb, False, black, lw, lc)


class Circle(ObjectMixin, physics.Circle):
    _init_physics = physics.Circle.__init__

    def draw_shape(self, screen):
        if self._color is not None:
            color = self._color
            lw, lc = self._line_width, self._line_color
            screen.draw_circle(self.bounding_box, True, color, lw, lc)

        elif self._line_width:
            lw, lc = self._line_width, self._line_color
            screen.draw_circle(self.bounding_box, False, black, lw, lc)


class Poly(ObjectMixin, physics.Poly):
    _init_physics = physics.Poly.__init__

    def draw_shape(self, screen):
        if self._color is not None:
            color = self._color
            lw, lc = self._line_width, self._line_color
            screen.draw_poly(self.bounding_box, True, color, lw, lc)

        elif self._line_width:
            lw, lc = self._line_width, self._line_color
            screen.draw_poly(self.bounding_box, False, black, lw, lc)


class Rectangle(ObjectMixin, Poly, physics.Rectangle):
    _init_physics = physics.Rectangle.__init__


class RegularPoly(ObjectMixin, Poly, physics.RegularPoly):
    _init_physics = physics.RegularPoly.__init__


if __name__ == '__main__':
    x = AABB(shape=(100, 200), world=set())
    type(x)
    print(x.mass)



import os
try:
    import PIL
except ImportError:
    PIL = None


class TextureManager(object):
    EXTENSIONS = ['png', 'gif', 'bmp', 'jpg', 'jpeg']

    def __init__(self, path=None):
        self._images = {}
        self._pillow = {}
        if path is None:
            path = get_default_image_path()
        self._path = os.path.abspath(path)

    def save_image(self, name, image):
        self._images[name] = image

    def get_image(self, name):
        try:
            return self._images[name]
        except KeyError:
            path = self.find_image_path(name)
            self._images[name] = img = self.image_from_path(path)
            return img

    def discard_image(self, name):
        if name in self._images:
            del self._images[name]
        if name in self._pillow:
            del self._pillow[name]

    def clear(self):
        self._images.clear()
        self._pillow.clear()

    def find_image_path(self, name):
        return os.path.join(self._path, name)

    # Funções para subclasses #################################################
    def image_from_path(self, path):
        img = PIL.Image.open(path)
        return self.import_image(img)

    def import_image(self, img):
        return img

    def export_image(self, img):
        return img

    # Transformações ##########################################################
    # Utilizam a biblioteca pillow, caso não exista suporte nativo às funções
    # utilizadas na biblioteca
    def mirror(self, name_or_img, horizontal=True):
        pass

    def slice(self, name_or_img, rect):
        pass

    def rescale(self, name_or_img, factor):
        pass

    def irotate(self, name_or_img, factor):
        pass

    def invert(self, name_or_img):
        pass

    # Informações sobre a imagem ##############################################
    def get_shape(self, img):
        raise NotImplementedError


###############################################################################
#                                Funções úteis
###############################################################################
def get_default_image_path():
    '''Retorna o caminho para a pasta /images/ abaixo do caminho do script
    principal em execução'''
    #script_path = sys.modules['__main__'].__file__
    #base_path = os.path.split(script_path)[0]
    #path = os.path.join(base_path, 'images')
    return ''


class PyGameTextureManager(TextureManager):

    def image_from_path(self, path):
        import pygame
        return pygame.image.load(path)

    def get_shape(self, img):
        x, y, width, height = img.get_rect()
        return width, height



#-*- coding: utf8 -*-
import os
import sys
import setuptools
from setuptools import setup

#
# Read VERSION from file and write it in the appropriate places
#
AUTHOR = 'Fábio Macêdo Mendes'
BASE, _ = os.path.split(__file__)
with open(os.path.join(BASE, 'VERSION')) as F:
    VERSION = F.read().strip()
with open(os.path.join(BASE, 'src', 'FGAme', 'meta.py'), 'w') as F:
    F.write(
        '# Auto-generated file. Please do not edit\n'
        '__version__ = %r\n' % VERSION +
        '__author__ = %r\n' % AUTHOR)

VERSION_BIG = VERSION.rpartition('.')[0]
IS_PYPY = 'PyPy' in sys.version
setup_kwds = {}

#
# Choose the default Python3 branch or the code converted by 3to2
#
PYSRC = 'src' if sys.version.startswith('3') else 'py2src'


#
# Find and configure extensions
#
def find_pyx():
    '''Return a list of all Cython scripts'''

    # Find files
    out = []
    for base, dirs, files in os.walk(os.path.join(os.getcwd(), 'src')):
        out.extend([os.path.join(base, f)
                    for f in files if f.endswith('.pyx')])

    # Remove cwd from path
    N = len(os.getcwd()) + 5
    out = [f[N:] for f in out]

    return out


def find_pure():
    '''Return a list of all .pxd files related to pure-python scripts'''

    # Find files
    out = []
    for base, dirs, files in os.walk(os.path.join(os.getcwd(), 'src')):
        out.extend([os.path.join(base, f)
                    for f in files if f.endswith('.pxd')])

    # Remove non-compilable files
    aux = []
    for f in out:
        with open(f) as F:
            line = F.readline()
        if line.startswith('#:'):
            if 'no-compile' in line:
                continue
        if os.path.exists(f[:-4] + '.py'):
            aux.append(f)
    out = aux

    # Remove cwd from path
    N = len(os.getcwd()) + 5
    out = [f[N:-4] + '.py' for f in out]

    return out


def get_extensions():
    '''Uses information on all pure-python scripts and all Cython scripts
    to create the extension objects'''

    from distutils.extension import Extension
    win_platforms = ['win32', 'cygwin']

    if IS_PYPY:
        return []

    exts = []
    for path in find_pyx() + find_pure():
        base, fname = os.path.split(path)
        ext_name = os.path.splitext(path)[0].replace(os.path.sep, '.')
        includes = ['src/%s' % base]
        if base != 'mathtools':
            includes.append('src/mathtools')

        ext = Extension(
            ext_name,
            ["src/%s" % path],
            libraries=(
                [] if sys.platform in win_platforms else ['m']),
            include_dirs=includes)

        exts.append(ext)

    return exts

#
# Configure environment
#

# Test if installation can compile extensions and configure them
# Currently only cpython accepts extensions
try:
    from Cython.Distutils import build_ext
except ImportError:
    # Ignore missing cython in alternative implementation
    if not (sys.platform.startswith('java') or sys.platform.startswith(
            'cli') or 'PyPy' in sys.version):
        raise
else:
    extensions = get_extensions()
    setup_kwds.update(
        cmdclass={
            "build_ext": build_ext},
        ext_modules=get_extensions())

#
# Main configuration script
#
setup(
    name='FGAme',
    version=VERSION,
    description='A game engine for 2D physics',
    author='Fábio Macêdo Mendes',
    author_email='fabiomacedomendes@gmail.com',
    url='https://github.com/fabiommendes/FGAme',
    long_description=open(os.path.join(BASE, 'README.txt')).read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
    ],

    package_dir={'': PYSRC},
    packages=setuptools.find_packages(PYSRC),
    license='GPL',

    install_requires=[
        'pysdl2>=0.5', # or 'pygame>=1.9.*'
        'pillow',
        'six',
        'smallshapes>=%s' % VERSION_BIG,
        'smallvectors>=%s' % VERSION_BIG,
        'pygeneric>=0.1.1',
    ],
    # dependency_links=[
    #    'http://pygame.org/ftp/',
    #    'http://www.lfd.uci.edu/~gohlke/pythonlibs/',
    #],
    **setup_kwds
)



from .base import *
#from .cocos_base_actions import *
#from .cocos_interval_actions import *
#from .cocos_instant_actions import *
# from .basegrid_actions import *
# from .quadmoveby_actions import *
# from .tiledgrid_actions import *
# from .grid3d_actions import *
# from .camera_actions import *
# from .move_actions import *
